# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import numpy as np
from typing import Tuple

from ...constants import SMALL, TWOPI


def newtone(ecc: float, e0: float) -> Tuple[float, float]:
    """Solves for the mean anomaly and true anomaly given the eccentric, parabolic, or
    hyperbolic anomalies.

    References:
        Vallado: 2022, p. 78, Algorithm 6

    Args:
        ecc (float): Eccentricity
        e0 (float): Eccentric anomaly in radians (-2pi to 2pi)

    Returns:
        tuple: (m, nu)
            m (float): Mean anomaly in radians (0 to 2pi)
            nu (float): True anomaly in radians (0 to 2pi)
    """
    # Circular orbit case - values are same as eccentric anomaly
    if abs(ecc) < SMALL:
        return e0, e0

    # Non-circular cases
    if ecc < 0.999:
        # Elliptical orbit
        m = e0 - ecc * np.sin(e0)
        sinv = (np.sqrt(1 - ecc**2) * np.sin(e0)) / (1 - ecc * np.cos(e0))
        cosv = (np.cos(e0) - ecc) / (1 - ecc * np.cos(e0))
        nu = np.arctan2(sinv, cosv)
    elif ecc > 1.0001:
        # Hyperbolic orbit
        m = ecc * np.sinh(e0) - e0
        sinv = (np.sqrt(ecc**2 - 1) * np.sinh(e0)) / (1 - ecc * np.cosh(e0))
        cosv = (np.cosh(e0) - ecc) / (1 - ecc * np.cosh(e0))
        nu = np.arctan2(sinv, cosv)
    else:
        # Parabolic orbit
        m = e0 + (1 / 3) * e0**3
        nu = 2 * np.arctan(e0)

    return m, nu


def newtonnu(
    ecc: float, nu: float, parabolic_lim_deg: float = 168
) -> Tuple[float, float]:
    """Solves for the eccentric anomaly and mean anomaly given the true anomaly.

    This function solves Kepler's equation when the true anomaly is known. The mean and
    eccentric, parabolic, or hyperbolic anomaly is also found. The default parabolic
    limit at 168 deg is arbitrary. The hyperbolic anomaly is also limited. The
    hyperbolic sine is used because it's not double-valued.

    References:
        Vallado: 2022, p. 78, Algorithm 5

    Args:
        ecc (float): Eccentricity of the orbit
        nu (float): True anomaly in radians
        parabolic_lim_deg (float, optional): The paraboloic limit in degrees
                                             (default is 168)

    Returns:
        tuple: (e0, m)
            e0 (float): Eccentric anomaly in radians
            m (float): Mean anomaly in radians
    """
    e0, m = np.inf, np.inf

    # Circular case
    if abs(ecc) < SMALL:
        e0, m = nu, nu
    # Elliptical case
    elif ecc < 1 - SMALL:
        sine = (np.sqrt(1 - ecc**2) * np.sin(nu)) / (1 + ecc * np.cos(nu))
        cose = (ecc + np.cos(nu)) / (1 + ecc * np.cos(nu))
        e0 = np.arctan2(sine, cose)
        m = e0 - ecc * np.sin(e0)
    # Hyperbolic case
    elif ecc > 1 + SMALL:
        if ecc > 1 and abs(nu) < np.pi - np.arccos(1 / ecc):
            sine = (np.sqrt(ecc**2 - 1) * np.sin(nu)) / (1 + ecc * np.cos(nu))
            e0 = np.arcsinh(sine)
            m = ecc * np.sinh(e0) - e0
    # Parabolic case
    else:
        if abs(nu) < np.radians(parabolic_lim_deg):
            e0 = np.tan(nu / 2)
            m = e0 + (e0**3) / 3

    # Update eccentric and mean anomaly to be within (0, 2pi) range
    if ecc < 1:
        m = np.fmod(m, TWOPI)
        if m < 0:
            m += TWOPI
        e0 = np.fmod(e0, TWOPI)

    return e0, m


def newtonm(ecc: float, m: float, n_iter: int = 50) -> Tuple[float, float]:
    """Solves for the eccentric anomaly and true anomaly given the mean anomaly using
    Newton-Raphson iteration.

    References:
        Vallado: 2022, p. 65, Algorithm 2
        Oltrogge: JAS 2015

    Args:
        ecc (float): Eccentricity of the orbit
        m (float): Mean anomaly in radians
        n_iter (int): Number of iterations for eccentric anomaly solving

    Returns:
        tuple: (e0, nu)
            e0 (float): Eccentric anomaly in radians
            nu (float): True anomaly in radians
    """
    # Define eccentricity thresholds
    # TODO: better definition/notes
    ecc_thresh_mid, ecc_thresh_high = 1.6, 3.6

    # Hyperbolic orbit
    if (ecc - 1) > SMALL:
        if ecc < ecc_thresh_mid:
            if (0 > m > -np.pi) or (m > np.pi):
                e0 = m - ecc
            else:
                e0 = m + ecc
        else:
            if ecc < ecc_thresh_high and abs(m) > np.pi:
                e0 = m - np.sign(m) * ecc
            else:
                e0 = m / (ecc - 1)

        e1 = e0 + ((m - ecc * np.sinh(e0) + e0) / (ecc * np.cosh(e0) - 1))
        ktr = 1
        while abs(e1 - e0) > SMALL and ktr <= n_iter:
            e0 = e1
            e1 = e0 + (m - ecc * np.sinh(e0) + e0) / (ecc * np.cosh(e0) - 1)
            ktr += 1

        sinv = -(np.sqrt(ecc**2 - 1) * np.sinh(e1)) / (1 - ecc * np.cosh(e1))
        cosv = (np.cosh(e1) - ecc) / (1 - ecc * np.cosh(e1))
        nu = np.arctan2(sinv, cosv)

    # Parabolic orbit
    elif abs(ecc - 1) < SMALL:
        s = 0.5 * (np.pi * 0.5 - np.arctan(1.5 * m))
        w = np.arctan(np.tan(s) ** (1 / 3))
        e0 = 2 / np.tan(2 * w)
        nu = 2 * np.arctan(e0)

    # Elliptical orbit
    elif ecc > SMALL:
        if (0 > m > -np.pi) or (m > np.pi):
            e0 = m - ecc
        else:
            e0 = m + ecc
        e1 = e0 + (m - e0 + ecc * np.sin(e0)) / (1 - ecc * np.cos(e0))
        ktr = 1
        while abs(e1 - e0) > SMALL and ktr <= n_iter:
            e0 = e1
            e1 = e0 + (m - e0 + ecc * np.sin(e0)) / (1 - ecc * np.cos(e0))
            ktr += 1

        sinv = (np.sqrt(1 - ecc**2) * np.sin(e1)) / (1 - ecc * np.cos(e1))
        cosv = (np.cos(e1) - ecc) / (1 - ecc * np.cos(e1))
        nu = np.arctan2(sinv, cosv)

    # Circular orbit
    else:
        nu, e0 = m, m

    return e0, nu



import numpy as np

from ... import constants as const


# TODO: Move to a more general location


def time_of_flight(a: float) -> float:
    """Calculate the time of flight for a given semi-major axis.

    Args:
        a (float): Semi-major axis of the orbit in km

    Returns:
        float: Time of flight in seconds
    """
    return np.pi * np.sqrt(a**3 / const.MU)


def specific_mech_energy(a: float) -> float:
    """Computes specific mechanical energy at a given semi-major axis.

    Args:
        a (float): Semi-major axis of the orbit in km

    Returns:
        float: Specific mechanical energy
    """
    return -const.MU / (2 * a)


def velocity_mag(r: float, a: float) -> float:
    """Computes velocity magnitude at a given radius and specific mechanical energy.

    Args:
        r (float): Radius of the orbit in km
        a (float): Semi-major axis of the orbit in km

    Returns:
        float: Velocity magnitude in km/s
    """
    return np.sqrt(2 * ((const.MU / r) + specific_mech_energy(a)))


def angular_velocity(a: float) -> float:
    """Computes the angular velocity at a given semi-major axis.

    Args:
        a (float): Semi-major axis of the orbit in km

    Returns:
        float: Angular velocity in rad/s
    """
    return np.sqrt(const.MU / a**3)


def deltav(v1: float, v2: float, theta: float) -> float:
    """Computes the delta-v for a given two velocity magnitudes and an angle.

    Args:
        v1 (float): Initial velocity in km/s
        v2 (float): Final velocity in km/s
        theta (float): Angle in radians

    Returns:
        float: Delta-v in km/s
    """
    return np.sqrt(v1**2 + v2**2 - 2 * v1 * v2 * np.cos(theta))


def semimajor_axis(r: float, e: float, nu: float) -> float:
    """Computes the semi-major axis for a given radius, eccentricity, and true anomaly.

    Args:
        r (float): Radius of the orbit in km
        e (float): Eccentricity of the orbit
        nu (float): True anomaly in radians

    Returns:
        float: Semi-major axis in km
    """
    return (r * (1 + e * np.cos(nu))) / (1 - e**2)


def period(a: float) -> float:
    """Computes the period of an orbit given the semi-major axis.

    Args:
        a (float): Semi-major axis of the orbit in km

    Returns:
        float: Period of the orbit in seconds
    """
    return const.TWOPI * np.sqrt(a**3 / const.MU)


def lowuz(z: float) -> float:
    """Computes the control parameter u using Chebyshev polynomial approximation.

    References:
        Alfano: "Optimal Many-Revolution Orbit Transfer", Journal of Guidance, Vol 8,
                 No. 1, 1985, pp. 155-157.

    Args:
        z (float): Input parameter

    Returns:
        u (float): Computed control parameter (bounded between 0.000001 and 0.999999)
    """
    # Assign coefficients and starting values
    # fmt: off
    alpha = np.array([
        2.467410607, -1.907470562, 35.892442177, -214.672979624, 947.773273608,
        -2114.861134906, 2271.240058672, -1127.457440108, 192.953875268, 8.577733773
    ])

    beta = np.array([
        0.4609698838, 13.7756315324, -69.1245316678, 279.0671832500, -397.6628952136,
        -70.0139935047, 528.0334266841, -324.9303836520, 20.5838245170, 18.8165370778
    ])
    # fmt: on

    # Compute u using Chebyshev polynomial approximation
    alphasum, betasum, zterm = 0, 1, 1
    for i in range(10):
        zterm *= z
        alphasum += zterm * alpha[i]
        betasum += zterm * beta[i]

    u = abs(alphasum / betasum)

    # Clamp the value of u to be within the range [0.000001, 0.999999]
    u = max(0.000001, min(u, 0.999999))

    return u



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 10 Oct 2019
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import math
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ... import constants as const


def legpolyn(
    latgc: float, order: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Computes Legendre polynomials for the gravity field.

    References:
        Vallado: 2022, p. 600-601, Eq. 8-56

    Args:
        latgc: Geocentric latitude of the satellite in radians (-pi to pi)
        order: Size of the gravity field (1 to ~170)

    Returns:
        tuple: (legarr_mu, legarr_gu, legarr_mn, legarr_gn)
            legarr_mu (np.ndarray): Montenbruck approach Legendre polynomials
            legarr_gu (np.ndarray): GTDS approach Legendre polynomials
            legarr_mn (np.ndarray): Normalized Montenbruck polynomials
            legarr_gn (np.ndarray): Normalized GTDS polynomials

    Notes:
        - Some recursions at high degree tesseral terms experience error for resonant
          orbits - these are valid for normalized and unnormalized expressions, as long
          as the remaining equations are consistent.
        - For satellite operations, orders up to about 120 are valid.
    """
    legarr_mu = np.zeros((order + 1, order + 1))
    legarr_gu = np.zeros((order + 1, order + 1))
    legarr_mn = np.zeros((order + 1, order + 1))
    legarr_gn = np.zeros((order + 1, order + 1))

    # Perform recursions (Montenbruck approach)
    legarr_mu[:2, :2] = [[1, 0], [np.sin(latgc), np.cos(latgc)]]

    # Legendre functions, zonal
    for n in range(2, order + 1):
        legarr_mu[n, n] = (2 * n - 1) * legarr_mu[1, 1] * legarr_mu[n - 1, n - 1]

    # Associated Legendre functions
    for n in range(2, order + 1):
        for m in range(n):
            if n == m + 1:
                legarr_mu[n, m] = (2 * m + 1) * legarr_mu[1, 0] * legarr_mu[m, m]
            else:
                legarr_mu[n, m] = (1 / (n - m)) * (
                    (2 * n - 1) * legarr_mu[1, 0] * legarr_mu[n - 1, m]
                    - (n + m - 1) * legarr_mu[n - 2, m]
                )

    # Normalize the Legendre polynomials
    for n in range(order + 1):
        for m in range(n + 1):
            factor = 1 if m == 0 else 2
            conv = np.sqrt(
                (math.factorial(n - m) * factor * (2 * n + 1)) / math.factorial(n + m)
            )
            legarr_mn[n, m] = conv * legarr_mu[n, m]

    # Perform recursions (GTDS approach)
    legarr_gu[:2, :2] = [[1, 0], [np.sin(latgc), np.cos(latgc)]]

    for n in range(2, order + 1):
        for m in range(n + 1):
            legarr_gu[n, m] = 0

    for n in range(2, order + 1):
        for m in range(n + 1):
            # Legendre functions, zonal
            if m == 0:
                legarr_gu[n, m] = (
                    (2 * n - 1) * legarr_gu[1, 0] * legarr_gu[n - 1, m]
                    - (n - 1) * legarr_gu[n - 2, m]
                ) / n
            else:
                # Associated Legendre functions
                if m == n:
                    legarr_gu[n, m] = (
                        (2 * n - 1) * legarr_gu[1, 1] * legarr_gu[n - 1, m - 1]
                    )
                else:
                    legarr_gu[n, m] = (
                        legarr_gu[n - 2, m]
                        + (2 * n - 1) * legarr_gu[1, 1] * legarr_gu[n - 1, m - 1]
                    )

    # Normalize the Legendre polynomials
    for n in range(order + 1):
        for m in range(n + 1):
            factor = 1 if m == 0 else 2
            conv1 = np.sqrt(
                (math.factorial(n - m) * factor * (2 * n + 1)) / math.factorial(n + m)
            )
            legarr_gn[n, m] = conv1 * legarr_gu[n, m]

    return legarr_mu, legarr_gu, legarr_mn, legarr_gn


def trigpoly(
    recef: ArrayLike, latgc: float, lon: float, order: int
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Computes accumulated Legendre polynomials and trigonometric terms.

    References:
        Vallado: 2022, p. 600-602

    Args:
        recef (array_like): ECEF satellite position vector in km
        latgc (float): Geocentric latitude of the satellite in radians
        lon (float): Longitude of the satellite in radians
        order (int): Size of the gravity field (1.. 2160..)

    Returns:
        tuple: (trig_arr, v_arr, w_arr)
            trig_arr (np.ndarray): Array of trigonometric terms
            v_arr (np.ndarray): V array of trigonometric terms
            w_arr (np.ndarray): W array of trigonometric terms
    """
    magr = np.linalg.norm(recef)
    l_ = 0

    # Initialize arrays
    trig_arr = np.zeros((order + 1, 3))
    v_arr = np.zeros((order + 2, order + 2))
    w_arr = np.zeros((order + 2, order + 2))

    # Trigonometric terms (GTDS approach)
    trig_arr[0, 0] = 0  # sin terms
    trig_arr[0, 1] = 1  # cos terms
    tlon = np.tan(latgc)
    trig_arr[1, 0] = np.sin(lon)
    trig_arr[1, 1] = np.cos(lon)
    clon = np.cos(lon)

    for m in range(2, order + 1):
        # Sine terms
        trig_arr[m, 0] = 2 * clon * trig_arr[m - 1, 0] - trig_arr[m - 2, 0]
        # Cosine terms
        trig_arr[m, 1] = 2 * clon * trig_arr[m - 1, 1] - trig_arr[m - 2, 1]
        # Tangent terms
        trig_arr[m, 2] = (m - 1) * tlon + tlon

    # Montenbruck approach for V and W arrays
    temp = const.RE / (magr * magr)
    v_arr[0, 0] = const.RE / magr
    v_arr[1, 0] = v_arr[0, 0] ** 2 * np.sin(latgc)

    for l_ in range(2, order + 2):
        x1 = ((2 * l_ - 1) / l_) * recef[1] * temp
        x2 = ((l_ - 1) / l_) * temp * const.RE
        v_arr[l_, 0] = x1 * v_arr[l_ - 1, 0] - x2 * v_arr[l_ - 2, 0]

    # Tesseral and sectoral values for L = m
    for l_ in range(1, order + 2):
        m = l_
        x1 = (2 * m - 1) * recef[0] * temp
        x2 = recef[1] * temp
        v_arr[l_, m] = x1 * v_arr[l_ - 1, m - 1] - x2 * w_arr[l_ - 1, m - 1]
        w_arr[l_, m] = x1 * w_arr[l_ - 1, m - 1] - x2 * v_arr[l_ - 1, m - 1]

    for m in range(l_ + 1, order + 1):
        if m <= order:
            x = (2 * l_ - 1) / (l_ - m) * recef[1] * temp
            v_arr[l_ + 1, m] = x * v_arr[l_, m]
            w_arr[l_ + 1, m] = x * w_arr[l_, m]

        for l2 in range(m + 2, order + 2):
            x1 = ((2 * l2 - 1) / (l2 - m)) * recef[1] * temp
            x2 = ((l2 + m - 1) / (l2 - m)) * temp * const.RE
            v_arr[l2, m] = x1 * v_arr[l2 - 1, m] - x2 * v_arr[l2 - 2, m]
            w_arr[l2, m] = x1 * w_arr[l2 - 1, m] - x2 * w_arr[l2 - 2, m]

    return trig_arr, v_arr, w_arr



# -----------------------------------------------------------------------------
# Name: constants.py
# Author: David Vallado
# Date: 2 Apr 2007
#
# Copyright (c) 2024
# For license information, see LICENSE file
# -----------------------------------------------------------------------------

import numpy as np

###############################################################################
# Mathematical Operations
###############################################################################

SMALL = 1e-10

# Distances
KM2M = 1e3
FT2M = 0.3048
MILE2M = 1609.344
NM2M = 1852
MILE2FT = 5280
MILEPH2KMPH = 0.44704
NMPH2KMPH = 0.5144444

# Time
DAY2SEC = 86400
DAY2MIN = 1440
DAY2HR = 24
HR2SEC = 3600
MIN2SEC = 60
YR2DAY = 365.25
CENT2YR = 100
CENT2DAY = CENT2YR * YR2DAY

# Angles
HALFPI = np.pi / 2
TWOPI = 2 * np.pi
DEG2MIN = 60
DEG2ARCSEC = DEG2MIN * MIN2SEC
ARCSEC2RAD = np.radians(1 / DEG2ARCSEC)
DEG2SEC = np.degrees(TWOPI) / DAY2SEC
DEG2HR = np.degrees(TWOPI) / DAY2HR
HR2RAD = DEG2HR * np.radians(1)

###############################################################################
# Astrodynamic Operations
###############################################################################

# Time
J2000 = 2451545  # Julian date of the epoch J2000.0 (noon)
J2000_UTC = 2451544.5  # Julian date of the epoch J2000.0 in UTC (midnight)
JD_TO_MJD_OFFSET = 2400000.5  # offset between Julian and Modified Julian dates

# EGM-08 (Earth) constants used here
# fmt: off
RE = 6378.1363                      # km
FLAT = 1 / 298.257223563
EARTHROT = 7.292115e-5              # rad/s
MU = 398600.4415                    # km^3/s^2
MUM = 3.986004415e14                # m^3/s^2
J2 = 0.001082626174
J4 = -1.6198976e-06
# fmt: on

# Derived constants from the base values

# Approximate Earth rotation
EARTHROT_APPROX = TWOPI / DAY2SEC  # rad/s

# Earth eccentricity
ECCEARTH = np.sqrt(2 * FLAT - FLAT**2)
ECCEARTHSQRD = ECCEARTH**2

# Earth radius
RENM = RE / NM2M
REFT = RE * 1e3 / FT2M

# Orbital period
TUSEC = np.sqrt(RE**3 / MU)
TUMIN = TUSEC / MIN2SEC
TUDAY = TUSEC / DAY2SEC
TUDAYSID = TUSEC / 86164.090524

# Earth rotation & rotational angular velocity
OMEGAARTHPTU = EARTHROT * TUSEC
OMEGAARTHPMIN = EARTHROT * MIN2SEC

# Orbital velocity
VELKPS = np.sqrt(MU / RE)
VELFPS = VELKPS * 1e3 / FT2M
VELPDMIN = VELKPS * MIN2SEC / RE
DEGSEC = (180 / np.pi) / TUSEC
RADPDAY = TWOPI * 1.002737909350795

# Astronomical distances & measurements
# fmt: off
SPEEDOFLIGHT = 299792.458           # km/s
AU2KM = 149597870.7                 # km
EARTH2MOON = 384400                 # km
MOONRADIUS = 1738                   # km
SUNRADIUS = 696000                  # km
# fmt: on

# Masses in kg
MASSSUN = 1.9891e30
MASSEARTH = 5.9742e24
MASSMOON = 7.3483e22

# Standard gravitational parameters in km^3/s^2
MUSUN = 1.32712428e11
MUMOON = 4902.799

# Obliquities
OBLIQUITYEARTH = np.radians(23.439291)



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 6 Aug 2008
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from dataclasses import dataclass
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ... import constants as const
from ..time.data import IAU80Array
from ..time.frame_conversions import eci2ecef
from ..twobody.frame_conversions import rv2razel, rv2tradec
from ..twobody.kepler import kepler
from ..twobody.utils import site
from .utils import finite_diff


@dataclass
class ObservationData:
    obstype: np.ndarray = None
    time: np.ndarray = None
    timef: np.ndarray = None
    latgd: np.ndarray = None
    lon: np.ndarray = None
    alt: np.ndarray = None
    jdut1: np.ndarray = None
    ttt: np.ndarray = None
    xp: np.ndarray = None
    yp: np.ndarray = None
    rng: np.ndarray = None
    az: np.ndarray = None
    el: np.ndarray = None
    trtasc: np.ndarray = None
    tdecl: np.ndarray = None
    noise_rng: np.ndarray = None
    noise_az: np.ndarray = None
    noise_el: np.ndarray = None
    noise_trtasc: np.ndarray = None
    noise_tdecl: np.ndarray = None


def findatwaatwb(
    iau80arr: IAU80Array,
    firstobs: int,
    lastobs: int,
    obsrecarr: ObservationData,
    percentchg: float,
    deltaamtchg: float,
    xnom: ArrayLike,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, float, float, float]:
    """Find the a and b matrices for the differential correction problem.

    References:
        Vallado: 2022, p. 782-785

    Args:
        iau80arr (IAU80Array): IAU 1980 data for nutation
        firstobs (int): Index of the first observation
        lastobs (int): Index of the last observation
        obsrecarr (ObservationData): Observation data records
        percentchg (float): Amount to modify the vectors in finite differencing
        deltaamtchg (float): Tolerance for small value in finite differencing
        xnom (np.ndarray): State vector in km and km/s

    Returns:
        tuple (atwa, atwb, atw, b, drng2, daz2, del2):
            atwa (np.ndarray): atwa matrix in km and rad
            atwb (np.ndarray): atwb matrix in km and rad
            atw (np.ndarray): atw matrix in km and rad
            b (np.ndarray): Matrix of observation residuals in km and rad
            drng2 (float): Range residual squared in km^2
            daz2 (float): Azimuth residual squared in rad^2
            del2 (float): Elevation residual squared in rad^2

    Notes:
        - It isn't critical for the propagations to use the highest fidelity techniques
        because we're only trying to find the "slope."

    TODO:
        - Verify units of outputs
        - This function is likely stale and will need to be updated
    """

    def adjust_angle(angle):
        if abs(angle) > np.pi:
            angle -= np.sign(angle) * const.TWOPI
        return angle

    # Initialize variables
    statesize = len(xnom)
    atwa = np.zeros((statesize, statesize))
    atwb = np.zeros((statesize, 1))
    atw = np.zeros((statesize, 3))
    a = np.zeros((3, statesize))
    b = np.zeros((3, 1))
    drng2 = daz2 = del2 = 0

    # Get observation index
    indobs = 1 if obsrecarr.obstype[0] == 0 else 3 if obsrecarr.obstype[0] == 2 else 2

    # Loop through all observations
    aznom = elnom = trtascnom = tdeclnom = 0
    for obsktr in range(firstobs - 1, lastobs):
        # Propagate the nominal vector to the epoch time
        dtsec = (
            obsrecarr.time[obsktr]
            - obsrecarr.time[0]
            + obsrecarr.timef[obsktr]
            - obsrecarr.timef[0]
        ) * const.DAY2SEC
        rnom, vnom = xnom[:3], xnom[3:]
        reci1, veci1 = kepler(rnom, vnom, dtsec)

        # Convert ECI to ECEF
        aeci = np.zeros(3)
        lod = ddpsi = ddeps = 0
        recef1, vecef1, _ = eci2ecef(
            reci1,
            veci1,
            aeci,
            obsrecarr.ttt[obsktr],
            obsrecarr.jdut1[obsktr],
            lod,
            obsrecarr.xp[obsktr],
            obsrecarr.yp[obsktr],
            ddpsi,
            ddeps,
            iau80arr,
        )

        # Get site vectors
        rsecef, vsecef = site(
            obsrecarr.latgd[obsktr], obsrecarr.lon[obsktr], obsrecarr.alt[obsktr]
        )

        # Get range, azimuth, and elevation (or right ascension and declination)
        if obsrecarr.obstype[obsktr] != 3:
            rngnom, aznom, elnom, *_ = rv2razel(
                recef1,
                vecef1,
                obsrecarr.latgd[obsktr],
                obsrecarr.lon[obsktr],
                obsrecarr.alt[obsktr],
            )
        else:
            rngnom, trtascnom, tdeclnom, *_ = rv2tradec(recef1, vecef1, rsecef, vsecef)

        # Calculate b matrix
        if obsrecarr.obstype[obsktr] == 0:
            b[0, 0] = obsrecarr.rng[obsktr] - rngnom
        elif obsrecarr.obstype[obsktr] == 1:
            b[0, 0] = obsrecarr.az[obsktr] - aznom
            b[0, 0] = adjust_angle(b[0, 0])
            b[1, 0] = obsrecarr.el[obsktr] - elnom
        elif obsrecarr.obstype[obsktr] == 2:
            b[0, 0] = obsrecarr.rng[obsktr] - rngnom
            b[1, 0] = obsrecarr.az[obsktr] - aznom
            b[1, 0] = adjust_angle(b[1, 0])
            b[2, 0] = obsrecarr.el[obsktr] - elnom
        else:
            b[0, 0] = obsrecarr.trtasc[obsktr] - trtascnom
            b[0, 0] = adjust_angle(b[0, 0])
            b[1, 0] = obsrecarr.tdecl[obsktr] - tdeclnom

        # Perturb each element in the state (elements or vectors)
        for j in range(statesize):
            # Use finite differencing to perturb state
            deltaamt, xnomp = finite_diff(j, percentchg, deltaamtchg, xnom)

            # Propagate the perturbed vector to the epoch time
            rnomp, vnomp = xnomp[:3], xnomp[3:]
            reci3, veci3 = kepler(rnomp, vnomp, dtsec)

            # Convert ECI to ECEF
            recef3, vecef3, _ = eci2ecef(
                reci3,
                veci3,
                aeci,
                obsrecarr.ttt[obsktr],
                obsrecarr.jdut1[obsktr],
                lod,
                obsrecarr.xp[obsktr],
                obsrecarr.yp[obsktr],
                ddpsi,
                ddeps,
                iau80arr,
            )

            # Get range, azimuth, and elevation (or right ascension and declination)
            rngpert = azpert = elpert = trtascpert = tdeclpert = 0
            if obsrecarr.obstype[obsktr] == 3:
                trrpert, trtascpert, tdeclpert, *_ = rv2tradec(
                    recef3, vecef3, rsecef, vsecef
                )
            else:
                rngpert, azpert, elpert, *_ = rv2razel(
                    recef3,
                    vecef3,
                    obsrecarr.latgd[obsktr],
                    obsrecarr.lon[obsktr],
                    obsrecarr.alt[obsktr],
                )

            # Calculate a matrix
            if obsrecarr.obstype[obsktr] == 0:
                a[0, j] = (rngpert - rngnom) / deltaamt
            elif obsrecarr.obstype[obsktr] == 1:
                a[0, j] = (azpert - aznom) / deltaamt
                a[1, j] = (elpert - elnom) / deltaamt
            elif obsrecarr.obstype[obsktr] == 2:
                a[0, j] = (rngpert - rngnom) / deltaamt
                a[1, j] = (azpert - aznom) / deltaamt
                a[2, j] = (elpert - elnom) / deltaamt
            else:
                a[0, j] = (trtascpert - trtascnom) / deltaamt
                a[1, j] = (tdeclpert - tdeclnom) / deltaamt

        # Form matrix combinations
        at = a.T

        # Assign weights
        w2 = w3 = weight = 1
        if obsrecarr.obstype[obsktr] == 0:
            w1 = 1 / (obsrecarr.noise_rng[obsktr] ** 2)
            drng2 += b[0, 0] ** 2 * w1
        elif obsrecarr.obstype[obsktr] == 1:
            w1 = 1 / (obsrecarr.noise_az[obsktr] ** 2)
            w2 = 1 / (obsrecarr.noise_el[obsktr] ** 2)
            daz2 += b[0, 0] ** 2 * w1
            del2 += b[1, 0] ** 2 * w2
        elif obsrecarr.obstype[obsktr] == 2:
            w1 = 1 / (obsrecarr.noise_rng[obsktr] ** 2)
            w2 = 1 / (obsrecarr.noise_az[obsktr] ** 2)
            w3 = 1 / (obsrecarr.noise_el[obsktr] ** 2)
            drng2 += b[0, 0] ** 2 * w1
            daz2 += b[1, 0] ** 2 * w2
            del2 += b[2, 0] ** 2 * w3
        else:
            w1 = 1 / (obsrecarr.noise_trtasc[obsktr] ** 2)
            w2 = 1 / (obsrecarr.noise_tdecl[obsktr] ** 2)
            drng2 += b[0, 0] ** 2 * w1
            daz2 += b[1, 0] ** 2 * w2

        # Form the atw matrix
        atw = np.zeros((statesize, indobs))
        for rowc in range(statesize):
            for colc in range(indobs):
                if colc == 0:
                    weight = w1
                elif colc == 1:
                    weight = w2
                elif colc == 2:
                    weight = w3
                atw[rowc, colc] = at[rowc, colc] * weight

        # Form the atwa and atwb matrices
        atwaacc, atwbacc = np.dot(atw, a), np.dot(atw, b)
        atwa += atwaacc
        atwb[:, 0] += atwbacc[:, 0]

    return atwa, atwb, atw, b, drng2, daz2, del2



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from typing import Tuple

import numpy as np

from ... import constants as const
from ...mathtime.vector import unit
from ..twobody.frame_conversions import rv2ntw
from ..twobody.kepler import kepler
from . import utils as utils


########################################################################################
# Coplanar Transfers
########################################################################################


def hohmann(
    rinit: float,
    rfinal: float,
    einit: float,
    efinal: float,
    nuinit: float,
    nufinal: float,
) -> Tuple[float, float, float]:
    """Calculates the delta-v values for a Hohmann transfer, either circle-to-circle
    or ellipse-to-ellipse.

    References:
        Vallado 2022, p. 326-329, Algorithm 36

    Args:
        rinit (float): Initial position magnitude in km
        rfinal (float): Final position magnitude in km
        einit (float): Eccentricity of the initial orbit
        efinal (float): Eccentricity of the final orbit
        nuinit (float): True anomaly of the initial orbit in radians (0 or pi)
        nufinal (float): True anomaly of the final orbit in radians (0 or pi)

    Returns:
        tuple: (deltava, deltavb, dtsec)
            deltava (float): Change in velocity at point A in km/s
            deltavb (float): Change in velocity at point B in km/s
            dtsec (float): Time of flight for the transfer in seconds
    """
    # Semi-major axes of initial, transfer, and final orbits
    ainit = utils.semimajor_axis(rinit, einit, nuinit)
    atran = (rinit + rfinal) / 2
    afinal = utils.semimajor_axis(rfinal, efinal, nufinal)

    # Initialize outputs
    deltava, deltavb, dtsec = 0, 0, 0

    if einit < 1 or efinal < 1:
        # Calculate delta-v at point A
        vinit = utils.velocity_mag(rinit, ainit)
        vtrana = utils.velocity_mag(rinit, atran)
        deltava = np.abs(vtrana - vinit)

        # Calculate delta-v at point B
        vfinal = utils.velocity_mag(rfinal, afinal)
        vtranb = utils.velocity_mag(rfinal, atran)
        deltavb = np.abs(vfinal - vtranb)

        # Calculate transfer time of flight
        dtsec = utils.time_of_flight(atran)

    return deltava, deltavb, dtsec


def bielliptic(
    rinit: float,
    rb: float,
    rfinal: float,
    einit: float,
    efinal: float,
    nuinit: float,
    nufinal: float,
) -> Tuple[float, float, float, float]:
    """Calculates the delta-v values for a bi-elliptic transfer, either circle-to-circle
    or ellipse-to-ellipse.

    References:
        Vallado 2022, pp. 326-330, Algorithm 37

    Args:
        rinit (float): Initial position magnitude in km
        rb (float): Intermediate position magnitude in km
        rfinal (float): Final position magnitude in km
        einit (float): Eccentricity of the initial orbit
        efinal (float): Eccentricity of the final orbit
        nuinit (float): True anomaly of the initial orbit in radians (0 or pi)
        nufinal (float): True anomaly of the final orbit in radians (0 or pi)

    Returns:
        tuple: (deltava, deltavb, deltavc, dtsec)
            deltava (float): Change in velocity at point A in km/s
            deltavb (float): Change in velocity at point B in km/s
            deltavc (float): Change in velocity at point C in km/s
            dtsec (float): Time of flight for the transfer in seconds
    """
    # Semi-major axes of initial, transfer, and final orbits
    ainit = utils.semimajor_axis(rinit, einit, nuinit)
    atran1 = (rinit + rb) * 0.5
    atran2 = (rb + rfinal) * 0.5
    afinal = utils.semimajor_axis(rfinal, efinal, nufinal)

    # Initialize outputs
    deltava, deltavb, deltavc, dtsec = 0, 0, 0, 0

    # Check if inputs represent elliptical orbits
    if einit < 1 and efinal < 1:
        # Calculate delta-v at point A
        vinit = utils.velocity_mag(rinit, ainit)
        vtran1a = utils.velocity_mag(rinit, atran1)
        deltava = np.abs(vtran1a - vinit)

        # Calculate delta-v at point B
        vtran1b = utils.velocity_mag(rb, atran1)
        vtran2b = utils.velocity_mag(rb, atran2)
        deltavb = np.abs(vtran1b - vtran2b)

        # Calculate delta-v at point C
        vtran2c = utils.velocity_mag(rfinal, atran2)
        vfinal = utils.velocity_mag(rfinal, afinal)
        deltavc = np.abs(vfinal - vtran2c)

        # Calculate total time of flight
        dtsec = utils.time_of_flight(atran1) + utils.time_of_flight(atran2)

    return deltava, deltavb, deltavc, dtsec


def one_tangent(
    rinit: float,
    rfinal: float,
    efinal: float,
    nuinit: float,
    nutran: float,
    tol: float = 1e-6,
) -> Tuple[float, float, float, float, float, float, float]:
    """Calculates the delta-v values for a one tangent transfer, either circle-to-circle
    or ellipse-to-ellipse.

    References:
        Vallado 2022, p. 335-338, Algorithm 38

    Args:
        rinit (float): Initial position magnitude in km
        rfinal (float): Final position magnitude in km
        efinal (float): Eccentricity of the final orbit
        nuinit (float): True anomaly of the initial orbit in radians (0 or pi)
        nutran (float): True anomaly of the transfer orbit in radians
                        (same quadrant as `nuinit`)
        tol (float): Tolerance for checking if transfer orbit is valid (default 1e-6)

    Returns:
        tuple: (deltava, deltavb, dtsec, etran, atran, vtrana, vtranb)
            deltava (float): Change in velocity at point A in km/s
            deltavb (float): Change in velocity at point B in km/s
            dtsec (float): Time of flight for the transfer in seconds
            etran (float): Eccentricity of the transfer orbit
            atran (float): Semi-major axis of the transfer orbit in km
            vtrana (float): Velocity of the transfer orbit at point A in km/s
            vtranb (float): Velocity of the transfer orbit at point B in km/s

    Raises:
        ValueError: If the one-tangent burn is not possible for the given inputs
    """
    # Initialize transfer time
    dtsec = 0

    # Ratio of initial to final orbit radii
    ratio = rinit / rfinal

    # Determine eccentricity of transfer orbit
    if abs(nuinit) < 0.01:  # near 0 or 180 degrees
        etran = (ratio - 1) / (np.cos(nutran) - ratio)
    else:
        etran = (ratio - 1) / (np.cos(nutran) + ratio)

    # Check if transfer orbit is valid
    if etran >= 0:
        # Semi-major axes of initial, final, and transfer orbits
        afinal = utils.semimajor_axis(rfinal, efinal, nutran)

        if abs(etran - 1) > tol:
            atran = utils.semimajor_axis(rinit, etran, nuinit)
        else:
            atran = np.inf  # parabolic orbit (infinite semi-major axis)

        # Calculate delta-V at point A
        vinit = np.sqrt(const.MU / rinit)
        vtrana = utils.velocity_mag(rinit, atran)
        deltava = np.abs(vtrana - vinit)

        # Calculate delta-V at point B
        vfinal = utils.velocity_mag(rfinal, afinal)
        vtranb = utils.velocity_mag(rfinal, atran)
        fpatranb = np.arctan2(etran * np.sin(nutran), 1 + etran * np.cos(nutran))
        fpafinal = np.arctan2(efinal * np.sin(nutran), 1 + efinal * np.cos(nutran))
        deltavb = utils.deltav(vfinal, vtranb, fpatranb - fpafinal)

        # Calculate time of flight
        if etran < 1:
            sinv = (np.sqrt(1 - etran**2) * np.sin(nutran)) / (
                1 + etran * np.cos(nutran)
            )
            cosv = (etran + np.cos(nutran)) / (1 + etran * np.cos(nutran))
            e = np.arctan2(sinv, cosv)
            eainit = 0 if abs(nuinit) < 0.01 else np.pi
            dtsec = np.sqrt(atran**3 / const.MU) * (
                e - etran * np.sin(e) - (eainit - etran * np.sin(eainit))
            )
    else:
        raise ValueError("The one-tangent burn is not possible for this case.")

    return deltava, deltavb, dtsec, etran, atran, vtrana, vtranb


########################################################################################
# Non-Coplanar Transfers
########################################################################################


def incl_only(deltai: float, vinit: float, fpa: float) -> float:
    """Calculates the delta-v for a change in inclination only.

    References:
        Vallado 2022, p. 346-348, Algorithm 39

    Args:
        deltai (float): Change in inclination in radians
        vinit (float): Initial velocity in km/s
        fpa (float): Flight path angle in radians

    Returns:
        float: Delta-v required for inclination change in km/s

    Notes:
        - Units are flexible for `vinit` and the output will match its units
    """
    return 2 * vinit * np.cos(fpa) * np.sin(0.5 * deltai)


def node_only(
    iinit: float,
    ecc: float,
    deltaraan: float,
    vinit: float,
    fpa: float,
    incl: float,
    tol: float = 1e-7,
) -> Tuple[float, float, float, float]:
    """Calculates the delta-v for a change in longitude of the ascending node.

    References:
        Vallado 2022, pp. 349-351, Algorithm 40

    Args:
        iinit (float): Initial inclination in radians
        ecc (float): Eccentricity of the initial orbit
        deltaraan (float): Change in right ascension of the ascending node in radians
        vinit (float): Initial velocity in km/s
        fpa (float): Flight path angle in radians
        incl (float): Inclination in radians
        tol (float): Tolerance for checking if transfer is elliptical (default 1e-7)

    Returns:
        tuple: (ifinal, deltav, arglat_init, arglat_final)
            ifinal (float): Final inclination in radians
            deltav (float): Change in velocity in km/s
            arglat_init (float): Initial argument of latitude in radians
            arglat_final (float): Final argument of latitude in radians

    Notes:
        - Units are flexible for `vinit` and `deltav` will match its units
    """
    if ecc > tol:
        # Elliptical orbit
        theta = np.arctan(np.sin(iinit) * np.tan(deltaraan))
        ifinal = np.arcsin(np.sin(theta) / np.sin(deltaraan))
        deltav = 2 * vinit * np.cos(fpa) * np.sin(0.5 * theta)

        # Initial argument of latitude
        arglat_init = np.pi / 2  # set at 90 degrees

    else:
        # Circular orbit
        ifinal = incl
        theta = np.arccos(np.cos(iinit) ** 2 + np.sin(iinit) ** 2 * np.cos(deltaraan))
        deltav = 2 * vinit * np.sin(0.5 * theta)

        # Initial argument of latitude
        arglat_init = np.arccos(
            (np.tan(iinit) * (np.cos(deltaraan) - np.cos(theta))) / np.sin(theta)
        )

    # Final argument of latitude
    arglat_final = np.arccos(
        (np.cos(incl) * np.sin(incl) * (1 - np.cos(deltaraan))) / np.sin(theta)
    )

    return ifinal, deltav, arglat_init, arglat_final


def incl_and_node(
    iinit: float, ifinal: float, deltaraan: float, vinit: float, fpa: float
) -> Tuple[float, float, float]:
    """Calculates the delta-v for a change in inclination and right ascension of the
    ascending node.

    References:
        Vallado 2022, p. 352, Algorithm 41

    Args:
        iinit (float): Initial inclination in radians
        ifinal (float): Final inclination in radians
        deltaraan (float): Change in right ascension of the ascending node in radians
        vinit (float): Initial velocity in km/s
        fpa (float): Flight path angle in radians

    Returns:
        tuple: (deltav, arglat_init, arglat_final)
            deltav (float): Change in velocity in km/s
            arglat_init (float): Initial argument of latitude in radians
            arglat_final (float): Final argument of latitude in radians

    Notes:
        - Units are flexible for `vinit` and `deltav` will match its units
    """
    # Pre-compute trigonometric values for efficiency
    cosdraan = np.cos(deltaraan)
    sinii, cosii = np.sin(iinit), np.cos(iinit)
    sinif, cosif = np.sin(ifinal), np.cos(ifinal)

    # Calculate theta
    cost = cosii * cosif + sinii * sinif * cosdraan
    theta = np.arccos(cost)
    sint = np.sin(theta)

    # Calculate delta-v
    deltav = incl_only(theta, vinit, fpa)

    # Calculate argument of latitude changes
    arglat_init = np.arccos((sinif * cosdraan - cost * sinii) / (sint * cosii))
    arglat_final = np.arccos((cosii * sinif - sinii * cosif * cosdraan) / sint)

    return deltav, arglat_init, arglat_final


########################################################################################
# Combined Transfers
########################################################################################


def min_combined(
    rinit: float,
    rfinal: float,
    einit: float,
    efinal: float,
    nuinit: float,
    nufinal: float,
    iinit: float,
    ifinal: float,
    use_optimal: bool = True,
    tol: float = 1e-6,
) -> Tuple[float, float, float, float, float]:
    """Calculates the delta-v and inclination change for the minimum velocity change
    between two non-coplanar orbits.

    References:
        Vallado 2022, p. 355-357, Algorithm 42

    Args:
        rinit (float): Initial position magnitude in km
        rfinal (float): Final position magnitude in km
        einit (float): Eccentricity of the initial orbit
        efinal (float): Eccentricity of the final orbit
        nuinit (float): True anomaly of the initial orbit in radians (0 or pi)
        nufinal (float): True anomaly of the final orbit in radians (0 or pi)
        iinit (float): Initial inclination in radians
        ifinal (float): Final inclination in radians
        use_optimal (bool): Use iterative optimization for inclination change
                            (default True)
        tol (float): Tolerance for inclination iteration (default 1e-6)

    Returns:
        tuple: (deltai_init, deltai_final, deltava, deltavb, dtsec)
            deltai_init (float): Inclination change at point A in radians
            deltai_final (float): Inclination change at point B in radians
            deltava (float): Delta-v at point A in km/s
            deltavb (float): Delta-v at point B in km/s
            dtsec (float): Time of flight for the transfer in seconds
    """
    # Compute semi-major axes
    a1 = (rinit * (1 + einit * np.cos(nuinit))) / (1 - einit**2)
    a2 = 0.5 * (rinit + rfinal)
    a3 = (rfinal * (1 + efinal * np.cos(nufinal))) / (1 - efinal**2)

    # Compute velocities
    vinit = utils.velocity_mag(rinit, a1)
    v1t = utils.velocity_mag(rinit, a2)
    vfinal = utils.velocity_mag(rfinal, a3)
    v3t = utils.velocity_mag(rfinal, a2)

    # Delta inclination
    tdi = ifinal - iinit

    # Delta-Vs
    temp = (1 / tdi) * np.arctan(np.sin(tdi) / ((rfinal / rinit) ** 1.5 + np.cos(tdi)))
    deltava = utils.deltav(v1t, vinit, temp * tdi)
    deltavb = utils.deltav(v3t, vfinal, tdi * (1 - temp))

    # Inclination change
    deltai_init = temp * tdi
    deltai_final = tdi * (1 - temp)

    # Compute transfer time of flight
    dtsec = utils.time_of_flight(a2)

    if not use_optimal:
        return deltai_init, deltai_final, deltava, deltavb, dtsec

    # Iterative optimization
    deltai_final_iter, n_iter = 100, 0
    while abs(deltai_init - deltai_final_iter) > tol:
        deltai_final_iter = deltai_init
        deltava = utils.deltav(v1t, vinit, deltai_final_iter)
        deltavb = utils.deltav(v3t, vfinal, tdi - deltai_final_iter)
        deltai_init = np.arcsin(
            (deltava * vfinal * v3t * np.sin(tdi - deltai_final_iter))
            / (vinit * v1t * deltavb)
        )
        n_iter += 1

    return deltai_init, tdi - deltai_init, deltava, deltavb, dtsec


def combined(
    rinit: float, rfinal: float, einit: float, nuinit: float, deltai: float
) -> Tuple[float, float, float, float, float, float, float]:
    """Calculates the delta-v for a Hohmann transfer between two orbits, considering
    inclination changes. The final orbit is assumed to be circular.

    References:
        Vallado 2007, p. 360-361, Example 6-7

    Args:
        rinit (float): Initial position magnitude in km
        rfinal (float): Final position magnitude in km
        einit (float): Eccentricity of the initial orbit
        nuinit (float): True anomaly of the initial orbit in radians (0 or pi)
        deltai (float): Inclination change in radians (final - initial)

    Returns:
        tuple: (deltai1, deltai2, deltava, deltavb, dtsec, gama, gamb)
            deltai1 (float): Inclination change at point A in radians
            deltai2 (float): Inclination change at point B in radians
            deltava (float): Delta-v at point A in km/s
            deltavb (float): Delta-v at point B in km/s
            dtsec (float): Time of flight for the transfer in seconds
            gama (float): Firing angle at point A in radians
            gamb (float): Firing angle at point B in radians

    TODO:
        - Support non-circular final orbits?
    """
    # Semi-major axes
    ainit = utils.semimajor_axis(rinit, einit, nuinit)
    atran = (rinit + rfinal) * 0.5

    # Velocities
    vinit = utils.velocity_mag(rinit, ainit)
    vtransa = utils.velocity_mag(rinit, atran)
    vfinal = np.sqrt(const.MU / rfinal)  # assumes circular orbit
    vtransb = utils.velocity_mag(rfinal, atran)

    # Proportions of inclination change
    ratio = rfinal / rinit
    s = (1 / deltai) * np.arctan(np.sin(deltai) / (ratio**1.5 + np.cos(deltai)))
    deltai1 = s * deltai
    deltai2 = (1 - s) * deltai

    # Delta-v calculations
    deltava = utils.deltav(vinit, vtransa, deltai1)
    deltavb = utils.deltav(vfinal, vtransb, deltai2)

    # Time of flight for the transfer
    dtsec = utils.time_of_flight(atran)

    # Firing angles
    gama = np.arccos(-(vinit**2 + deltava**2 - vtransa**2) / (2 * vinit * deltava))
    gamb = np.arccos(-(vtransb**2 + deltavb**2 - vfinal**2) / (2 * vtransb * deltavb))

    return deltai1, deltai2, deltava, deltavb, dtsec, gama, gamb


########################################################################################
# Rendezvous
########################################################################################


def rendezvous_coplanar(
    rcsint: float,
    rcstgt: float,
    phasei: float,
    einit: float,
    efinal: float,
    nuinit: float,
    nufinal: float,
    kint: int,
    ktgt: int,
    tol: float = 1e-6,
) -> Tuple[float, float, float]:
    """Calculates parameters for a Hohmann transfer rendezvous.

    References:
        Vallado 2022, pp. 361-367, Algorithms 44 and 45

    Args:
        rcsint (float): Radius of the interceptor's circular orbit in km
        rcstgt (float): Radius of the target's circular orbit in km
        phasei (float): Initial phase angle (target - interceptor) in radians
        einit (float): Eccentricity of the initial orbit
        efinal (float): Eccentricity of the final orbit
        nuinit (float): True anomaly of the initial orbit in radians (0 or pi)
        nufinal (float): True anomaly of the final orbit in radians (0 or pi)
        kint (int): Number of interceptor orbits
        ktgt (int): Number of target orbits to wait
        tol (float): Tolerance for checking if orbit is the same (default 1e-6 rad/s)

    Returns:
        tuple: (phasef, waittime, deltav)
            phasef (float): Final phase angle in radians
            waittime (float): Wait time before next intercept opportunity in seconds
            deltav (float): Total change in velocity in km/s
    """
    # Angular velocities
    angvelint = utils.angular_velocity(rcsint)
    angveltgt = utils.angular_velocity(rcstgt)
    vint = np.sqrt(const.MU / rcsint)

    # Same orbit case
    if abs(angvelint - angveltgt) < tol:
        periodtrans = (ktgt * const.TWOPI + phasei) / angveltgt
        atrans = (const.MU * (periodtrans / (const.TWOPI * kint)) ** 2) ** (1 / 3)

        # Check for intersection with Earth
        rp = 2 * atrans - rcsint
        if rp < 1:
            raise ValueError("Error: the transfer orbit intersects the Earth.")

        # Calculate delta-V
        vtrans = np.sqrt((2 * const.MU / rcsint) - (const.MU / atrans))
        deltav = 2 * (vtrans - vint)

        # Calculate final phase angle and wait time
        phasef = phasei
        waittime = periodtrans

    else:
        # Different orbits
        atrans = (rcsint + rcstgt) / 2
        dttutrans = np.pi * np.sqrt(atrans**3 / const.MU)

        # Calculate final phase angle
        leadang = angveltgt * dttutrans
        phasef = np.pi - leadang
        if phasef < 0:
            phasef += np.pi

        # Calculate wait time
        waittime = (phasef - phasei + const.TWOPI * ktgt) / (angvelint - angveltgt)

        # Semi-major axes
        a1 = utils.semimajor_axis(rcsint, einit, nuinit)
        a2 = (rcsint + rcstgt) / 2
        a3 = utils.semimajor_axis(rcstgt, efinal, nufinal)

        # Delta-V at point A
        vinit = utils.velocity_mag(rcsint, a1)
        vtransa = utils.velocity_mag(rcsint, a2)
        deltava = abs(vtransa - vinit)

        # Delta-V at point B
        vfinal = utils.velocity_mag(rcstgt, a3)
        vtransb = utils.velocity_mag(rcstgt, a2)
        deltavb = abs(vfinal - vtransb)

        # Total delta-V
        deltav = deltava + deltavb

    return phasef, waittime, deltav


def rendezvous_noncoplanar(
    phasei: float,
    aint: float,
    atgt: float,
    kint: int,
    ktgt: int,
    nodeint: float,
    truelon: float,
    deltai: float,
) -> Tuple[float, float, float, float, float, float]:
    """Calculates parameters for a non-coplanar Hohmann transfer maneuver.

    References:
        Vallado 2022, pp. 369-373, Algorithm 46

    Args:
        phasei (float): Initial phase angle in radians
        aint (float): Semi-major axis of interceptor orbit in km
        atgt (float): Semi-major axis of target orbit in km
        kint (int): Number of interceptor orbits
        ktgt (int): Number of target orbits
        nodeint (float): Longitude of the ascending node of the interceptor in radians
        truelon (float): True longitude of the target in radians
        deltai (float): Change in inclination in radians (final - initial)

    Returns:
        tuple: (ttrans, tphase, dvphase, dvtrans1, dvtrans2, aphase)
            ttrans (float): Transfer time in seconds
            tphase (float): Phase time in seconds
            dvphase (float): Delta-v for phasing in km/s
            dvtrans1 (float): Delta-v for first transfer in km/s
            dvtrans2 (float): Delta-v for second transfer in km/s
            aphase (float): Semi-major axis for phasing orbit in km/s
    """
    # Angular velocities
    angvelint = utils.angular_velocity(aint)
    angveltgt = utils.angular_velocity(atgt)

    # Calculate transfer time
    atrans = (aint + atgt) / 2
    ttrans = utils.time_of_flight(atrans)

    # Calculate phase time
    deltatnode = phasei / angvelint
    lead = angveltgt * ttrans
    omeganode = angveltgt * deltatnode
    phasenew = nodeint + np.pi - (truelon + omeganode)
    leadnew = np.pi + phasenew
    tphase = (leadnew - lead + const.TWOPI * ktgt) / angveltgt

    # Semi-major axis of phasing orbit
    aphase = (const.MU * (tphase / (const.TWOPI * kint)) ** 2) ** (1 / 3)

    # Calculate phasing delta-V
    vphase = utils.velocity_mag(aint, aphase)
    dvphase = vphase - np.sqrt(const.MU / aint)

    # Calculate delta-V for first transfer
    vtrans1 = utils.velocity_mag(aint, atrans)
    dvtrans1 = vtrans1 - vphase

    # Calculate delta-V for second transfer
    vtrans2 = utils.velocity_mag(atgt, atrans)
    vtgt = np.sqrt(const.MU / atgt)
    dvtrans2 = utils.deltav(vtgt, vtrans2, deltai)

    return ttrans, tphase, dvphase, dvtrans1, dvtrans2, aphase


########################################################################################
# Low Thrust
########################################################################################


def low_thrust(
    ainit: float,
    afinal: float,
    iinit: float,
    mdot: float,
    accelinit: float,
    lambda_: float,
    steps: int = 360,
) -> Tuple[float, float]:
    """Calculate the delta-V and time of flight for a low-thrust, circular transfer.

    References:
        Vallado 2022, p. 382-392, Algorithm 47

    Args:
        ainit (float): Initial semi-major axis in km
        afinal (float): Final semi-major axis in km
        iinit (float): Initial inclination in radians
        mdot (float): Mass flow rate in kg/s
        accelinit (float): Initial acceleration in km/s^2
        lambda_ (float): Control parameter for optimization
        steps (int, optional): Number of steps per orbit (defaults to 360)

    Returns:
        tuple: (deltav, tof)
            deltav (float): Total change in velocity in km/s
            tof (float): Time of flight in seconds
    """
    # Calculate ratio and transition SMA
    ratio = afinal / ainit
    atrans = (afinal + ainit) * 0.5  # km

    # Set up canonical units
    du = ainit
    duptu = np.sqrt(const.MU / ainit)
    atrans /= du  # DU

    # Initial state vector (assuming circular orbit)
    r1 = np.array([ainit, 0, 0])
    magr = np.linalg.norm(r1)
    v = np.sqrt(const.MU / magr)
    v1 = np.array([0, v * np.cos(iinit), v * np.sin(iinit)])

    # Get step size
    dtsec = utils.period(magr) / steps  # steps per orbit

    # Initialize variables
    # cosf is the angle from the node (perigee) - start at 0
    a1 = 1  # normalized to initial radius DU
    tof = rev = cosfold = 0
    nodevec = unit(r1)

    while a1 < (afinal / ainit):
        # Calculate steering angle and acceleration
        cv = 1 / (4 * lambda_**2 * a1**2 + 1)
        cosf = utils.lowuz(cv)
        steering = np.arctan(cosf / np.sqrt(1 / (1 / cv - 1)))
        accel = accelinit / (1 + mdot * tof)  # km/s^2

        # Calculate acceleration vector
        avec = accel * np.array([0, np.cos(steering), np.sin(steering)])

        # Convert to NTW frame
        rntw1, vntw1, tmntw = rv2ntw(r1, v1)
        acc1 = np.dot(tmntw.T, avec)  # NTW to ECI
        vtot = v1 + acc1 * dtsec

        # Propagate orbit to next step
        r1n, v1n = kepler(r1, vtot, dtsec)

        # Calculate the new node angle
        magr, magv = np.linalg.norm(r1n), np.linalg.norm(v1n)
        a1 = utils.specific_mech_energy((magv**2 * 0.5) - (const.MU / magr))
        cosf = np.dot(r1n, nodevec) / magr
        if cosf * cosfold < 0:
            rev += 0.5

        # Update variables
        tof += dtsec
        r1, v1 = r1n, v1n
        dtsec = utils.period(a1) / steps
        cosfold = cosf
        a1 /= du

    # Calculate delta-V
    deltav = (1 - np.sqrt(1 / ratio)) * duptu  # km/s

    return deltav, tof



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import numpy as np
from numpy.typing import ArrayLike
from typing import Tuple

from .utils import gibbs
from ... import constants as const


def calculate_time_intervals(
    jd1: float, jdf1: float, jd2: float, jdf2: float, jd3: float, jdf3: float
) -> Tuple[float, float, float]:
    """Calculate time intervals between three Julian dates.

    Args:
        jd1 (float): Julian date of first sighting (days from 4713 BC)
        jdf1 (float): Julian date fraction of first sighting (days from 4713 BC)
        jd2 (float): Julian date of second sighting (days from 4713 BC)
        jdf2 (float): Julian date fraction of second sighting (days from 4713 BC)
        jd3 (float): Julian date of third sighting (days from 4713 BC)
        jdf3 (float): Julian date fraction of third sighting (days from 4713 BC)

    Returns:
        tuple: (tau12, tau13, tau32)
            tau12 (float): Time interval between t1 and t2 in seconds
            tau13 (float): Time interval between t1 and t3 in seconds
            tau32 (float): Time interval between t3 and t2 in seconds
    """
    tau12 = ((jd1 - jd2) + (jdf1 - jdf2)) * const.DAY2SEC
    tau13 = ((jd1 - jd3) + (jdf1 - jdf3)) * const.DAY2SEC
    tau32 = ((jd3 - jd2) + (jdf3 - jdf2)) * const.DAY2SEC

    return tau12, tau13, tau32


def calculate_los_vectors(decl: list[float], rtasc: list[float]) -> list[np.ndarray]:
    """Calculate line-of-sight unit vectors from declination and right ascension.

    Args:
        decl (list[float]): List of declinations in radians
        rtasc (list[float]): List of right ascensions in radians

    Returns:
        list[np.ndarray]: List of line-of-sight unit vectors
    """
    return [
        np.array([np.cos(dec) * np.cos(ra), np.cos(dec) * np.sin(ra), np.sin(dec)])
        for dec, ra in zip(decl, rtasc)
    ]


def _halley_iteration(
    poly: np.ndarray,
    bigr2c_default: float = 20000 / const.RE,
    tol: float = 8e-5,
    max_iter: int = 15,
):
    """Perform Halley's iteration to refine the value of bigr2."""
    # Select the max real root (use fallback of ~GPS altitude if necessary)
    roots = np.roots(poly)
    real_roots = roots[np.isreal(roots)].real
    bigr2c = max(real_roots) if real_roots.size > 0 else bigr2c_default

    # Halley iteration for refining bigr2
    bigr2, kk = 100, 0
    while abs(bigr2 - bigr2c) > tol and kk < max_iter:
        bigr2 = bigr2c
        deriv = bigr2**8 + poly[2] * bigr2**6 + poly[5] * bigr2**3 + poly[8]
        deriv1 = 8 * bigr2**7 + 6 * poly[2] * bigr2**5 + 3 * poly[5] * bigr2**2
        deriv2 = 56 * bigr2**6 + 30 * poly[2] * bigr2**4 + 6 * poly[5] * bigr2
        bigr2c = bigr2 - (2 * deriv * deriv1) / (2 * deriv1**2 - deriv * deriv2)
        kk += 1

    return bigr2c


def laplace(
    decl1: float,
    decl2: float,
    decl3: float,
    rtasc1: float,
    rtasc2: float,
    rtasc3: float,
    jd1: float,
    jdf1: float,
    jd2: float,
    jdf2: float,
    jd3: float,
    jdf3: float,
    diffsites: bool,
    rseci1: ArrayLike,
    rseci2: ArrayLike,
    rseci3: ArrayLike,
) -> Tuple[np.ndarray, np.ndarray]:
    """Solve orbit determination problem using three optical sightings and Laplace's
    method.

    References:
        Vallado: 2022, p. 441-445

    Args:
        decl1 (float): Declination of first sighting in radians
        decl2 (float): Declination of second sighting in radians
        decl3 (float): Declination of third sighting in radians
        rtasc1 (float): Right ascension of first sighting in radians
        rtasc2 (float): Right ascension of second sighting in radians
        rtasc3 (float): Right ascension of third sighting in radians
        jd1 (float): Julian date of first sighting (days from 4713 BC)
        jdf1 (float): Julian date fraction of first sighting (days from 4713 BC)
        jd2 (float): Julian date of second sighting (days from 4713 BC)
        jdf2 (float): Julian date fraction of second sighting (days from 4713 BC)
        jd3 (float): Julian date of third sighting (days from 4713 BC)
        jdf3 (float): Julian date fraction of third sighting (days from 4713 BC)
        diffsites (bool): True if different sites, False if same site
        rseci1 (array_like): ECI site position vector of first sighting in km
        rseci2 (array_like):  ECI site position vector of second sighting in km
        rseci3 (array_like):  ECI site position vector of third sighting in km

    Returns:
        tuple: (r2, v2)
            r2 (np.ndarray): ECI position vector in km
            v2 (np.ndarray): ECI velocity vector in km/s
    """
    # Convert Julian dates to time intervals
    tau12, tau13, tau32 = calculate_time_intervals(jd1, jdf1, jd2, jdf2, jd3, jdf3)

    # Line-of-sight unit vectors
    los1, los2, los3 = calculate_los_vectors(
        [decl1, decl2, decl3], [rtasc1, rtasc2, rtasc3]
    )

    # Check denominators for zero to avoid division errors
    if any(tau == 0 for tau in (tau12, tau13, tau32)):
        raise ValueError(
            "One or more time intervals (tau12, tau13, tau32) are zero, causing"
            "division by zero."
        )

    # Normalize to canonical units
    tau12c = tau12 / const.TUSEC
    tau13c = tau13 / const.TUSEC
    tau32c = tau32 / const.TUSEC
    rseci1c = np.array(rseci1) / const.RE
    rseci2c = np.array(rseci2) / const.RE
    rseci3c = np.array(rseci3) / const.RE

    # Lagrange coefficients
    s1 = -tau32c / (tau12c * tau13c)
    s2 = (tau12c + tau32c) / (tau12c * tau32c)
    s3 = -tau12c / (-tau13c * tau32c)
    s4 = 2 / (tau12c * tau13c)
    s5 = 2 / (tau12c * tau32c)
    s6 = 2 / (-tau13c * tau32c)

    # First and second derivatives of los vectors
    ldot = s1 * los1 + s2 * los2 + s3 * los3
    lddot = s4 * los1 + s5 * los2 + s6 * los3

    # Derivatives of site position vectors
    if not diffsites:
        # Same sites
        earth_rot_vec = [0, 0, const.EARTHROT * const.TUSEC]  # canonical Earth rotation
        rs2dot = np.cross(earth_rot_vec, rseci2c)
        rs2ddot = np.cross(earth_rot_vec, rs2dot)
    else:
        # Different sites
        rs2dot = s1 * rseci1c + s2 * rseci2c + s3 * rseci3c
        rs2ddot = s4 * rseci1c + s5 * rseci2c + s6 * rseci3c

    # Compute determinants
    dmat = np.column_stack((los2, ldot, lddot))
    dmat1 = np.column_stack((los2, ldot, rs2ddot))
    dmat2 = np.column_stack((los2, ldot, rseci2c))
    dmat3 = np.column_stack((los2, rs2ddot, lddot))
    dmat4 = np.column_stack((los2, rseci2c, lddot))

    d = 2 * np.linalg.det(dmat)
    d1c = np.linalg.det(dmat1)
    d2c = np.linalg.det(dmat2)
    d3c = np.linalg.det(dmat3)
    d4c = np.linalg.det(dmat4)

    # Check determinant value
    if abs(d) < const.SMALL:
        raise ValueError("Determinant is too small; system may be singular.")

    # Solve the 8th-order polynomial
    l2dotrs = np.dot(los2, rseci2c)
    poly = np.zeros(9)
    poly[0] = 1
    poly[2] = l2dotrs * 4 * d1c / d - 4 * d1c**2 / d**2 - np.linalg.norm(rseci2c) ** 2
    poly[5] = l2dotrs * 4 * d2c / d - 8 * d1c * d2c / d**2
    poly[8] = -4 * d2c**2 / d**2

    # Halley iteration for refining bigr2c
    bigr2c = _halley_iteration(poly)

    # Solve for rho and rho dot
    rho = -2 * d1c / d - 2 * d2c / (bigr2c**3 * d)
    rhodot = -d3c / d - d4c / (bigr2c**3 * d)

    # Position and velocity vectors at the middle
    r2c = rho * los2 + rseci2c
    v2c = rhodot * los2 + rho * ldot + rs2dot

    return r2c * const.RE, v2c * const.RE / const.TUSEC


def _update_fg_series(
    ll: int,
    r1: np.ndarray,
    r2: np.ndarray,
    r3: np.ndarray,
    v2: np.ndarray,
    tau12: float,
    tau32: float,
    bigr2: float,
    ll_iters: int = 8,
) -> Tuple[float, float, float, float]:
    """Helper function to compute or refine the f and g series.

    Args:
        ll (int): Current iteration count
        r1 (np.ndarray): ECI position vector at t1 in km
        r2 (np.ndarray): ECI position vector at t2 in km
        r3 (np.ndarray): ECI position vector at t3 in km
        v2 (np.ndarray): ECI velocity vector at t2 in km/s
        tau12 (float): Time interval between t1 and t2 in seconds
        tau32 (float): Time interval between t3 and t2 in seconds
        bigr2 (float): Magnitude of r2 in km
        ll_iters (int, optional): Maximum number of iterations for exact method
                                  (default 8)

    Returns:
        tuple: (f1, g1, f3, g3)
            f1 (float): Value of f1
            g1 (float): Value of g1
            f3 (float): Value of f3
            g3 (float): Value of g3
    """
    if ll <= ll_iters:
        # Approximate method for early iterations
        u = const.MU / np.linalg.norm(r2) ** 3
        rdot = np.dot(r2, v2) / np.linalg.norm(r2)
        udot = -3 * const.MU * rdot / np.linalg.norm(r2) ** 4

        f1 = 1 - 0.5 * u * tau12**2 - (1 / 6) * udot * tau12**3
        g1 = tau12 - (1 / 6) * u * tau12**3 - (1 / 12) * udot * tau12**4
        f3 = 1 - 0.5 * u * tau32**2 - (1 / 6) * udot * tau32**3
        g3 = tau32 - (1 / 6) * u * tau32**3 - (1 / 12) * udot * tau32**4
    else:
        # Exact method for later iterations
        theta12 = np.arccos(
            np.clip(np.dot(r1, r2) / (np.linalg.norm(r1) * np.linalg.norm(r2)), -1, 1)
        )
        theta23 = np.arccos(
            np.clip(np.dot(r2, r3) / (np.linalg.norm(r2) * np.linalg.norm(r3)), -1, 1)
        )

        f1 = 1 - np.linalg.norm(r1) * (1 - np.cos(theta12)) / bigr2
        g1 = np.linalg.norm(r1) * np.linalg.norm(r2) * np.sin(theta12) / np.sqrt(bigr2)
        f3 = 1 - np.linalg.norm(r3) * (1 - np.cos(theta23)) / bigr2
        g3 = np.linalg.norm(r3) * np.linalg.norm(r2) * np.sin(theta23) / np.sqrt(bigr2)

    return f1, g1, f3, g3


def gauss(
    decl1: float,
    decl2: float,
    decl3: float,
    rtasc1: float,
    rtasc2: float,
    rtasc3: float,
    jd1: float,
    jdf1: float,
    jd2: float,
    jdf2: float,
    jd3: float,
    jdf3: float,
    rseci1: ArrayLike,
    rseci2: ArrayLike,
    rseci3: ArrayLike,
) -> Tuple[np.ndarray, np.ndarray]:
    """Solve orbit determination problem using three optical sightings and the
    Gaussian method.

    References:
        Vallado: 2022, p. 448, Algorithm 52

    Args:
        decl1 (float): Declination of first sighting in radians
        decl2 (float): Declination of second sighting in radians
        decl3 (float): Declination of third sighting in radians
        rtasc1 (float): Right ascension of first sighting in radians
        rtasc2 (float): Right ascension of second sighting in radians
        rtasc3 (float): Right ascension of third sighting in radians
        jd1 (float): Julian date of first sighting (days from 4713 BC)
        jdf1 (float): Julian date fraction of first sighting (days from 4713 BC)
        jd2 (float): Julian date of second sighting (days from 4713 BC)
        jdf2 (float): Julian date fraction of second sighting (days from 4713 BC)
        jd3 (float): Julian date of third sighting (days from 4713 BC)
        jdf3 (float): Julian date fraction of third sighting (days from 4713 BC)
        rseci1 (array_like): ECI site position vector of first sighting in km
        rseci2 (array_like): ECI site position vector of second sighting in km
        rseci3 (array_like): ECI site position vector of third sighting in km

    Returns:
        tuple: (r2, v2)
            r2 (np.ndarray): ECI position vector in km
            v2 (np.ndarray): ECI velocity vector in km/s
    """
    # Time intervals in seconds
    tau12, tau13, tau32 = calculate_time_intervals(jd1, jdf1, jd2, jdf2, jd3, jdf3)

    # Line-of-sight unit vectors
    los1, los2, los3 = calculate_los_vectors(
        [decl1, decl2, decl3], [rtasc1, rtasc2, rtasc3]
    )

    # Normalize to canonical units
    tau12c = tau12 / const.TUSEC
    tau32c = tau32 / const.TUSEC
    rseci1c = np.array(rseci1) / const.RE
    rseci2c = np.array(rseci2) / const.RE
    rseci3c = np.array(rseci3) / const.RE

    # Construct l-matrix and determinant
    lmat = np.column_stack([los1, los2, los3])
    d = np.linalg.det(lmat)

    # Check determinant value
    if abs(d) < const.SMALL:
        raise ValueError("Determinant too small; unable to proceed with calculations.")

    # Inverse of l-matrix
    lmati = np.linalg.inv(lmat)

    # Range-site matrix
    rsmatc = np.column_stack([rseci1c, rseci2c, rseci3c])
    lir = lmati @ rsmatc

    # Calculate coefficients for polynomial
    a1 = tau32c / (tau32c - tau12c)
    a1u = (tau32c * ((tau32c - tau12c) ** 2 - tau32c**2)) / (6 * (tau32c - tau12c))
    a3 = -tau12c / (tau32c - tau12c)
    a3u = -(tau12c * ((tau32c - tau12c) ** 2 - tau12c**2)) / (6 * (tau32c - tau12c))

    # Polynomial terms
    d1c = lir[1, 0] * a1 - lir[1, 1] + lir[1, 2] * a3
    d2c = lir[1, 0] * a1u + lir[1, 2] * a3u
    magrs2 = np.linalg.norm(rseci2c)
    l2dotrs = np.dot(los2, rseci2c)

    # Solve the 8th-order polynomial
    poly = np.zeros(9)
    poly[0] = 1
    poly[2] = -(d1c**2 + 2 * d1c * l2dotrs + magrs2**2)
    poly[5] = -2 * (l2dotrs * d2c + d1c * d2c)
    poly[8] = -(d2c**2)

    # Halley iteration for refining bigr2c
    bigr2c = _halley_iteration(poly)

    # Ensure a valid radius solution
    if bigr2c < 0 or bigr2c * const.RE > 50000:
        bigr2c = 35000 / const.RE  # adjust to a GEO-like altitude

    # Convert some values back to non-canonical units
    bigr2 = bigr2c * const.RE
    a1u *= const.TUSEC**2
    a3u *= const.TUSEC**2

    # Solve for improved estimates of f and g series
    u = const.MU / bigr2**3
    c1 = a1 + a1u * u
    c2 = -1
    c3 = a3 + a3u * u

    # Compute range values
    rsmat = np.column_stack([rseci1, rseci2, rseci3])
    lir = lmati @ rsmat
    cmat = np.array([-c1, -c2, -c3]).reshape(-1, 1)
    rhomat = lir @ cmat

    # Form position vectors
    r1 = rhomat[0, 0] * los1 / c1 + rseci1
    r2 = rhomat[1, 0] * los2 / c2 + rseci2
    r3 = rhomat[2, 0] * los3 / c3 + rseci3

    # Find middle velocity vector with gibbs
    v2, *_ = gibbs(r1, r2, r3)

    return r2, v2


def doubler_iter(
    magr1in: float,
    magr2in: float,
    los1: ArrayLike,
    los2: ArrayLike,
    los3: ArrayLike,
    rsite1: ArrayLike,
    rsite2: ArrayLike,
    rsite3: ArrayLike,
    tau12: float,
    tau32: float,
    n12: int,
    n13: int,
    n23: int,
) -> Tuple[np.ndarray, np.ndarray, float, float, float, float, float, float, float]:
    """Perform the iterative work for the double-r angles-only routine.

    References:
        Vallado: 2022, p. 449-452

    Args:
        magr1in (float): Magnitude of the first sighting position vector
        magr2in (float): Magnitude of the second sighting position vector
        los1 (array_like): Line-of-sight unit vector for the first sighting
        los2 (array_like): Line-of-sight unit vector for the second sighting
        los3 (array_like): Line-of-sight unit vector for the third sighting
        rsite1 (array_like): ECI site position vector of the first sighting
        rsite2 (array_like): ECI site position vector of the second sighting
        rsite3 (array_like): ECI site position vector of the third sighting
        tau12 (float): Time interval between t1 and t2 in seconds
        tau32 (float): Time interval between t3 and t2 in seconds
        n12 (int): Number of days between the first and second sightings
        n13 (int): Number of days between the first and third sightings
        n23 (int): Number of days between the second and third sightings

        Returns:
            tuple: (r2, r3, f1, f2, q1, magr1, magr2, a, deltae32)
                r2 (np.ndarray): ECI position vector at t2 in km
                r3 (np.ndarray): ECI position vector at t3 in km
                f1 (float): Value of f1 coefficient
                f2 (float): Value of f2 coefficient
                q1 (float): Quality estimate of the solution
                magr1 (float): Magnitude of the first sighting position vector
                magr2 (float): Magnitude of the second sighting position vector
                a (float): Semi-major axis of the orbit in km
                deltae32 (float): Eccentric anomaly difference between obs 3 and 2 in
                                  radians

    """
    # Define default range value for when the square root is negative
    # Use this because hyperbolic likely at shorter times, lower alt
    default_range = 300

    # Range coefficients
    cc1 = 2 * np.dot(los1, rsite1)
    cc2 = 2 * np.dot(los2, rsite2)

    # Magnitude of the site position vectors
    magrsite1 = np.linalg.norm(rsite1)
    magrsite2 = np.linalg.norm(rsite2)

    # Compute rho1 and rho2
    tempsq1 = cc1**2 - 4 * (magrsite1**2 - magr1in**2)
    tempsq1 = default_range if tempsq1 < 0 else tempsq1
    rho1 = (-cc1 + np.sqrt(tempsq1)) * 0.5
    tempsq2 = cc2**2 - 4 * (magrsite2**2 - magr2in**2)
    tempsq2 = default_range if tempsq2 < 0 else tempsq2
    rho2 = (-cc2 + np.sqrt(tempsq2)) * 0.5

    # Compute r1 and r2
    r1 = rho1 * np.array(los1) + np.array(rsite1)
    r2 = rho2 * np.array(los2) + np.array(rsite2)
    magr1 = np.linalg.norm(r1)
    magr2 = np.linalg.norm(r2)

    # Compute the cross product and determine rho3
    w = np.cross(r1, r2) / (magr1 * magr2)
    rho3 = -np.dot(rsite3, w) / np.dot(los3, w)
    r3 = rho3 * np.array(los3) + np.array(rsite3)
    magr3 = np.linalg.norm(r3)

    # Delta true anomaly between obs 2 and 1
    cosdv21 = np.dot(r2, r1) / (magr2 * magr1)
    sindv21 = np.linalg.norm(np.cross(r2, r1)) / (magr2 * magr1)
    dv21 = np.arctan2(sindv21, cosdv21) + const.TWOPI * n12

    # Delta true anomaly between obs 3 and 1
    cosdv31 = np.dot(r3, r1) / (magr3 * magr1)
    sindv31 = np.sqrt(1 - cosdv31**2)
    dv31 = np.arctan2(sindv31, cosdv31) + const.TWOPI * n13

    # Delta true anomaly between obs 3 and 2
    cosdv32 = np.dot(r3, r2) / (magr3 * magr2)
    sindv32 = np.linalg.norm(np.cross(r3, r2)) / (magr3 * magr2)

    # Compute the semi-parameter
    if dv31 > np.pi:
        p = (sindv21 - sindv31 + sindv32) / (
            -sindv31 / magr2 + sindv21 / magr3 + sindv32 / magr1
        )
    else:
        c1 = (magr1 * sindv31) / (magr2 * sindv32)
        c3 = (magr1 * sindv21) / (magr3 * sindv32)
        p = (c3 * magr3 - c1 * magr2 + magr1) / (-c1 + c3 + 1)

    # Compute the cosines of the eccentric anomalies
    ecosv1 = p / magr1 - 1
    ecosv2 = p / magr2 - 1
    ecosv3 = p / magr3 - 1

    # Compute the sines of the eccentric anomalies
    if dv21 != np.pi:
        esinv2 = (-cosdv21 * ecosv2 + ecosv1) / sindv21
    else:
        esinv2 = (cosdv32 * ecosv2 - ecosv3) / sindv32

    # Eccentricity and semi-major axis
    e = np.sqrt(ecosv2**2 + esinv2**2)
    a = p / (1 - e**2)

    # Compute the delta mean and eccentric anomalies
    deltam12 = 0
    if e**2 < 1:
        # Non-hyperbolic case
        n = np.sqrt(const.MU / a**3)
        s = magr2 / p * np.sqrt(1 - e**2) * esinv2
        c = magr2 / p * (e**2 + ecosv2)

        # Delta eccentric anomaly between obs 3 and 2
        sinde32 = magr3 / np.sqrt(a * p) * sindv32 - magr3 / p * (1 - cosdv32) * s
        cosde32 = 1 - magr2 * magr3 / (a * p) * (1 - cosdv32)
        deltae32 = np.arctan2(sinde32, cosde32) + const.TWOPI * n23

        # Delta eccentric anomaly between obs 1 and 2
        sinde21 = magr1 / np.sqrt(a * p) * sindv21 + magr1 / p * (1 - cosdv21) * s
        cosde21 = 1 - magr2 * magr1 / (a * p) * (1 - cosdv21)
        deltae21 = np.arctan2(sinde21, cosde21) + const.TWOPI * n12

        # Delta mean anomalies
        deltam32 = deltae32 + 2 * s * (np.sin(deltae32 / 2)) ** 2 - c * np.sin(deltae32)
        deltam12 = (
            -deltae21 + 2 * s * (np.sin(deltae21 / 2)) ** 2 + c * np.sin(deltae21)
        )
    else:
        # Hyperbolic case
        if a > 0:
            a = -a
            p = -p
        n = np.sqrt(const.MU / -(a**3))
        s = magr2 / p * np.sqrt(e**2 - 1) * esinv2
        c = magr2 / p * (e**2 + ecosv2)

        # Delta eccentric anomaly between obs 3 and 2
        sindh32 = magr3 / np.sqrt(-a * p) * sindv32 - magr3 / p * (1 - cosdv32) * s
        deltah32 = np.log(sindh32 + np.sqrt(sindh32**2 + 1))
        deltam32 = (
            -deltah32 + 2 * s * (np.sinh(deltah32 / 2)) ** 2 + c * np.sinh(deltah32)
        )
        deltae32 = deltah32  # fix to match MATLAB

    # Calculate the f coefficients and quality estimate
    f1 = tau12 - deltam12 / n
    f2 = tau32 - deltam32 / n
    q1 = np.sqrt(f1**2 + f2**2)

    return r2, r3, f1, f2, q1, magr1, magr2, a, deltae32


def doubler(
    decl1: float,
    decl2: float,
    decl3: float,
    rtasc1: float,
    rtasc2: float,
    rtasc3: float,
    jd1: float,
    jdf1: float,
    jd2: float,
    jdf2: float,
    jd3: float,
    jdf3: float,
    rsite1: ArrayLike,
    rsite2: ArrayLike,
    rsite3: ArrayLike,
    rng1: float,
    rng2: float,
    pctchg: float = 0.005,
    tol_km: float = 0.1,
    max_iterations: int = 15,
) -> Tuple[np.ndarray, np.ndarray]:
    """Solve orbit determination problem using the double-r technique.

    References:
        Vallado: 2022, p. 449-452, Algorithm 53

    Args:
        decl1 (float): Declination of first sighting in radians
        decl2 (float): Declination of second sighting in radians
        decl3 (float): Declination of third sighting in radians
        rtasc1 (float): Right ascension of first sighting in radians
        rtasc2 (float): Right ascension of second sighting in radians
        rtasc3 (float): Right ascension of third sighting in radians
        jd1 (float): Julian date of first sighting (days from 4713 BC)
        jdf1 (float): Julian date fraction of first sighting (days from 4713 BC)
        jd2 (float): Julian date of second sighting (days from 4713 BC)
        jdf2 (float): Julian date fraction of second sighting (days from 4713 BC)
        jd3 (float): Julian date of third sighting (days from 4713 BC)
        jdf3 (float): Julian date fraction of third sighting (days from 4713 BC)
        rsite1 (array_like): ECI site position vector of first sighting in km
        rsite2 (array_like): ECI site position vector of second sighting in km
        rsite3 (array_like): ECI site position vector of third sighting in km
        rng1 (float): Range to first sighting in km
        rng2 (float): Range to second sighting in km
        pctchg (float, optional): Percentage change for iterative method (default 0.005)
        tol_km (float, optional): Position tolerance for convergence in km (default 0.1)
        max_iterations (int, optional): Maximum number of iterations (default 15)

    Returns:
        tuple: (r2, v2)
            r2 (np.ndarray): ECI position vector in km
            v2 (np.ndarray): ECI velocity vector in km/s
    """

    def calculate_fcoeffs(magr1in, magr2in, deltar1=None, deltar2=None):
        """Helper function to calculate partial derivatives and updated deltas."""
        if deltar1:
            magr1in += deltar1
        if deltar2:
            magr2in += deltar2

        # Call doubler to compute intermediate values
        _, _, f1, f2, q, *_ = doubler_iter(
            magr1in,
            magr2in,
            los1,
            los2,
            los3,
            rsite1,
            rsite2,
            rsite3,
            tau12,
            tau32,
            n12,
            n13,
            n23,
        )
        return f1, f2, q

    # Time differences in seconds
    tau12, tau13, tau32 = calculate_time_intervals(jd1, jdf1, jd2, jdf2, jd3, jdf3)

    # Period in seconds (assumed 1 day for Earth)
    n12 = np.floor(abs(tau12 / const.DAY2SEC))
    n13 = np.floor(abs(tau13 / const.DAY2SEC))
    n23 = np.floor(abs((tau12 + tau32) / const.DAY2SEC))

    # Line-of-sight unit vectors
    los1, los2, los3 = calculate_los_vectors(
        [decl1, decl2, decl3], [rtasc1, rtasc2, rtasc3]
    )

    # Iterative variables (make sure newqr is < oldqr to start)
    magr1in, magr2in = rng1, rng2
    magr1old, magr2old, oldqr, newqr = np.inf, np.inf, np.inf, 1e10

    # Main loop to get three values of the double-r for processing
    ktr = 0
    while (
        (abs(magr1in - magr1old) > tol_km or abs(magr2in - magr2old) > tol_km)
        and ktr < max_iterations
        and newqr < oldqr
    ):
        ktr += 1
        magr1o, magr2o, oldqr = magr1in, magr2in, newqr

        # Compute nominal values
        f1, f2, q1 = calculate_fcoeffs(magr1in, magr2in)

        # Compute perturbations
        # Re-calculate f1 and f2 with r1 = r1 + delta r1
        deltar1, deltar2 = pctchg * magr1o, pctchg * magr2o
        f1delr1, f2delr1, q2 = calculate_fcoeffs(magr1o, magr2o, deltar1=deltar1)
        pf1pr1, pf2pr1 = (f1delr1 - f1) / deltar1, (f2delr1 - f2) / deltar1

        # Re-calculate f1 and f2 with r2 = r2 + delta r2
        f1delr2, f2delr2, q3 = calculate_fcoeffs(magr1o, magr2o, deltar2=deltar2)
        pf1pr2, pf2pr2 = (f1delr2 - f1) / deltar2, (f2delr2 - f2) / deltar2

        # Compute delta updates
        magr1in, magr2in = magr1o, magr2o
        delta = pf1pr1 * pf2pr2 - pf2pr1 * pf1pr2
        delta1 = pf2pr2 * f1 - pf1pr2 * f2
        delta2 = pf1pr1 * f2 - pf2pr1 * f1
        deltar1 = -delta1 / delta if abs(delta) > const.SMALL else -delta1
        deltar2 = -delta2 / delta if abs(delta) > const.SMALL else -delta2

        # Limit corrections to avoid overshooting
        chkamt = 0.15
        deltar1 = np.clip(deltar1, -chkamt * magr1in, chkamt * magr1in)
        deltar2 = np.clip(deltar2, -chkamt * magr2in, chkamt * magr2in)

        # Update magnitudes and quality metric
        magr1old, magr2old = magr1in, magr2in
        magr1in += deltar1
        magr2in += deltar2
        newqr = np.sqrt(q1**2 + q2**2 + q3**2)

        # Reduce percentage change for next iteration
        pctchg *= 0.5

    # Final calculation for updated r2 and v2
    r2, r3, _, _, _, _, magr2, a, deltae32 = doubler_iter(
        magr1in,
        magr2in,
        los1,
        los2,
        los3,
        rsite1,
        rsite2,
        rsite3,
        tau12,
        tau32,
        n12,
        n13,
        n23,
    )

    f = 1 - a / magr2 * (1 - np.cos(deltae32))
    g = tau32 - np.sqrt(a**3 / const.MU) * (deltae32 - np.sin(deltae32))
    v2 = (r3 - f * r2) / g

    return r2, v2



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import warnings
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ... import constants as const
from ...mathtime.vector import unit


def gibbs(
    r1: ArrayLike,
    r2: ArrayLike,
    r3: ArrayLike,
    tol_angle: float = np.radians(1),
    orbit_tol: float = 1e-6,
) -> Tuple[np.ndarray, float, float, float]:
    """Determine the velocity at the middle point of the 3 given position vectors using
    the Gibbs method.

    References:
        Vallado: 2022, p. 460-467, Algorithm 54

    Args:
        r1 (array_like): ECI position vector #1 in km
        r2 (array_like): ECI position vector #2 in km
        r3 (array_like): ECI position vector #3 in km
        tol_angle (float, optional): Tolerance for angles (1 degree)
        orbit_tol (float, optional): Tolerance for orbit calculations (default 1e-6)

    Returns:
        tuple: (v2, theta12, theta23, copa)
            v2 (np.ndarray): ECI velocity vector at r2 in km/s
            theta12 (float): Angle between r1 and r2 in radians
            theta23 (float): Angle between r2 and r3 in radians
            copa (float): Co-planarity angle in radians
    """
    # Initialize variables
    theta12, theta23 = 0, 0

    # Magnitudes of position vectors
    magr1 = np.linalg.norm(r1)
    magr2 = np.linalg.norm(r2)
    magr3 = np.linalg.norm(r3)

    # Initialize velocity vector
    v2 = np.zeros(3)

    # Cross products
    p = np.cross(r2, r3)
    q = np.cross(r3, r1)
    w = np.cross(r1, r2)

    # Co-planarity angle
    copa = np.arcsin(np.dot(unit(p), unit(r1)))

    # Check for coplanarity
    if abs(copa) > np.sin(tol_angle):
        warnings.warn(
            "Vectors are not coplanar - results might be inaccurate.", UserWarning
        )

    # Sum of cross products
    d = p + q + w
    magd = np.linalg.norm(d)

    # Weighted sum of position vectors
    n = magr1 * p + magr2 * q + magr3 * w
    magn = np.linalg.norm(n)

    # Check if orbit determination is possible
    # Both `d` and `n` must be in the same direction and non-zero
    if any(x < orbit_tol for x in [magd, magn, np.dot(unit(n), unit(d))]):
        warnings.warn("Orbit determination is not possible.", UserWarning)
        return v2, theta12, theta23, copa

    # Angles between position vectors
    theta12 = np.arccos(np.clip(np.dot(r1, r2) / (magr1 * magr2), -1, 1))
    theta23 = np.arccos(np.clip(np.dot(r2, r3) / (magr2 * magr3), -1, 1))

    # Differences in position vector magnitudes
    r1mr2 = magr1 - magr2
    r3mr1 = magr3 - magr1
    r2mr3 = magr2 - magr3

    # S vector
    s = r1mr2 * np.array(r3) + r3mr1 * np.array(r2) + r2mr3 * np.array(r1)

    # B vector
    b = np.cross(d, r2)

    # Scaling factor
    lg = np.sqrt(const.MU / (magd * magn))

    # Compute velocity at r2
    tover2 = lg / magr2
    v2 = tover2 * b + lg * s

    return v2, theta12, theta23, copa


def hgibbs(
    r1: ArrayLike,
    r2: ArrayLike,
    r3: ArrayLike,
    jd1: float,
    jd2: float,
    jd3: float,
    tol_angle: float = np.radians(1),
) -> Tuple[np.ndarray, float, float, float]:
    """Determines the velocity at the middle point of the 3 given position vectors
    using the Herrick-Gibbs method.

    References:
        Vallado: 2022, p. 467-472, Algorithm 55

    Args:
        r1 (array_like): ECI position vector #1 in km
        r2 (array_like): ECI position vector #2 in km
        r3 (array_likey): ECI position vector #3 in km
        jd1 (float): Julian date of the 1st sighting in days
        jd2 (float): Julian date of the 2nd sighting in days
        jd3 (float): Julian date of the 3rd sighting in days
        tol_angle (float, optional): Tolerance for angles in radians (default = 1 deg)

    Returns:
        tuple: (v2, theta12, theta23, copa)
            v2 (np.ndarray): Velocity vector at r2 in km/s
            theta12 (float): Angle between r1 and r2 in radians
            theta23 (float): Angle between r2 and r3 in radians
            copa (float): Co-planarity angle in radians
    """
    # Magnitudes of position vectors
    magr1 = np.linalg.norm(r1)
    magr2 = np.linalg.norm(r2)
    magr3 = np.linalg.norm(r3)

    # Time differences in seconds
    dt21 = (jd2 - jd1) * const.DAY2SEC
    dt31 = (jd3 - jd1) * const.DAY2SEC
    dt32 = (jd3 - jd2) * const.DAY2SEC

    # Calculate coplanarity
    p = np.cross(r2, r3)
    copa = np.arcsin(np.dot(unit(p), unit(r1)))

    # Check for coplanarity
    if abs(copa) > np.sin(tol_angle):
        warnings.warn(
            "Vectors are not coplanar - results might be inaccurate.", UserWarning
        )

    # Check angle tolerance between position vectors
    theta12 = np.arccos(np.clip(np.dot(r1, r2) / (magr1 * magr2), -1, 1))
    theta23 = np.arccos(np.clip(np.dot(r2, r3) / (magr2 * magr3), -1, 1))

    # Warn if angles exceed tolerance
    if theta12 > tol_angle or theta23 > tol_angle:
        warnings.warn("Angles between vectors exceed tolerance.", UserWarning)

    # Herrick-Gibbs method
    term1 = -dt32 * (1 / (dt21 * dt31) + const.MU / (12 * magr1**3))
    term2 = (dt32 - dt21) * (1 / (dt21 * dt32) + const.MU / (12 * magr2**3))
    term3 = dt21 * (1 / (dt32 * dt31) + const.MU / (12 * magr3**3))

    # Calculate velocity at r2
    v2 = term1 * np.array(r1) + term2 * np.array(r2) + term3 * np.array(r3)

    return v2, theta12, theta23, copa



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import calendar
import logging
from enum import Enum
from typing import Tuple

import numpy as np

from .calendar import days2mdh
from .utils import hms2sec, sec2hms
from .. import constants as const


# Initialize the logger
logger = logging.getLogger(__name__)


# Constants
JULIAN_DATE_REFERENCE_YEAR = 1900  # reference year for the Julian Date
JULIAN_DATE_1900 = 2415019.5  # Julian date for January 0, 1900
JULIAN_DATE_EPOCH_OFFSET = 4716  # Julian date offset from the Gregorian calendar
DST_RULE_CHANGE_YEAR = 2007  # year when DST rules changed in the US
DST_CHANGE_UTC_HOUR = -10  # hour when DST changes in UTC (02:00)


class CalendarType(Enum):
    JULIAN = "j"
    GREGORIAN = "g"


def jd2sse(julian_date: float) -> float:
    """Converts Julian Date to seconds since epoch.

    Args:
        julian_date (float): The Julian Date (days from 4713 BC)

    Returns:
        float: Seconds since epoch (1 Jan 2000 00:00:00 UTC)
    """
    return (julian_date - const.J2000_UTC) * const.DAY2SEC


def jday(
    year: int, month: int, day: int, hour: int = 0, minute: int = 0, second: float = 0.0
) -> Tuple[float, float]:
    """Calculate the Julian Date (JD) and fractional day.

    This function finds the Julian Date given the year, month, day, and time.

    References:
        Vallado: 2022, p. 184-185, Algorithm 14

    Args:
        year (int): Year (e.g., 2024)
        month (int): Month (1 to 12)
        day (int): Day (1 to 31)
        hour (int, optional): Universal Time hour (0 to 23) (defaults to 0)
        minute (int, optional): Universal Time minute (0 to 59) (defaults to 0)
        second (float, optional): Universal Time second (0.0 to 59.999)
                                  (defaults to 0.0)

    Returns:
        tuple: (jd, jd_frac)
            jd (float): Julian Date
            jd_frac (float): Fractional part of the Julian Date
    """
    # Calculate Julian Date
    jd = (
        367 * year
        - np.floor((7 * (year + np.floor((month + 9) / 12))) * 0.25)
        + np.floor(275 * month / 9)
        + day
        + 1721013.5
    )  # use - 678987.0 to go to MJD directly

    # Calculate fractional part of the day
    jd_frac = (second + minute * const.MIN2SEC + hour * const.HR2SEC) / const.DAY2SEC

    # Adjust if jd_frac > 1
    if jd_frac > 1:
        jd += np.floor(jd_frac)
        jd_frac -= np.floor(jd_frac)

    return jd, jd_frac


def jday_all(
    year: int,
    month: int,
    day: int,
    hour: int,
    minute: int,
    second: float,
    calendar_type: CalendarType = CalendarType.GREGORIAN,
) -> Tuple[float, float]:
    """Calculates the Julian Date and fractional Julian Day.

    The Julian Date is defined by each elapsed day since noon, Jan 1, 4713 BC.

    References:
        Vallado: 2022, p. 185

    Args:
        year (int): Year (e.g., 1900 .. 2100)
        month (int): Month (1 .. 12)
        day (int): Day (1 .. 31)
        hour (int): Universal Time hour (0 .. 23)
        minute (int): Universal Time minute (0 .. 59)
        second (float): Universal Time second (0.0 .. 59.999)
        calendar_type (CalendarType, optional): Calendar type (Julian or Gregorian)
                                                (Defaults to CalendarType.GREGORIAN)

    Returns:
        tuple: (jd, jdfrac)
            jd (float): Julian Date
            jdfrac (float): Fractional part of the Julian Date
    """
    if month <= 2:
        year -= 1
        month += 12

    # Determine B based on the calendar type
    if calendar_type == CalendarType.JULIAN:
        b = 0
    elif calendar_type == CalendarType.GREGORIAN:
        b = 2 - (year // 100) + (year // 400)
    else:
        raise ValueError(
            "Invalid calendar type. Must be either CalendarType.JULIAN or"
            "CalendarType.GREGORIAN"
        )

    # Compute Julian Date
    jd = (
        int(const.YR2DAY * (year + JULIAN_DATE_EPOCH_OFFSET))
        + int(30.6001 * (month + 1))
        + day
        + b
        - 1524.5
    )

    # Compute fractional day
    jdfrac = (hour * const.HR2SEC + minute * const.MIN2SEC + second) / const.DAY2SEC

    # Normalize jdfrac if it exceeds 1.0
    if jdfrac >= 1:
        jd += int(jdfrac)
        jdfrac %= 1

    return jd, jdfrac


def invjday(jd: float, jdfrac: float) -> Tuple[int, int, int, int, int, float]:
    """Converts Julian Date and fractional day to calendar date and time.

    References:
        Vallado: 2022, p. 203-204, Algorithm 22

    Args:
        jd (float): Julian Date (days from 4713 BC)
        jdfrac (float): Fractional part of the Julian Date

    Returns:
        tuple: (year, month, day, hour, minute, second)
            year (int): Year
            month (int): Month
            day (int): Day
            hour (int): Hour
            minute (int): Minute
            second (float): Second

    Notes:
        - This assumes the Gregorian calendar type.
    """
    # Normalize jdfrac if it spans multiple days
    if abs(jdfrac) >= 1:
        jd += int(jdfrac)
        jdfrac %= 1

    # Adjust for fraction of a day in the Julian Date
    dt = jd - int(jd) - 0.5
    if abs(dt) > 1e-8:
        jd -= dt
        jdfrac += dt

    # Compute year and day of year
    temp = jd - JULIAN_DATE_1900
    tu = temp / const.YR2DAY
    year = JULIAN_DATE_REFERENCE_YEAR + int(tu)
    leap_years = (year - (JULIAN_DATE_REFERENCE_YEAR + 1)) // 4
    days = int(
        temp
        - ((year - JULIAN_DATE_REFERENCE_YEAR) * np.floor(const.YR2DAY) + leap_years)
    )

    # Handle start-of-year edge case
    if days + jdfrac < 1:
        year -= 1
        leap_years = (year - (JULIAN_DATE_REFERENCE_YEAR + 1)) // 4
        days = int(
            temp
            - (
                (year - JULIAN_DATE_REFERENCE_YEAR) * np.floor(const.YR2DAY)
                + leap_years
            )
        )

    # Convert days of year + fractional day to calendar date and time
    month, day, hour, minute, second = days2mdh(year, days + jdfrac)

    return year, month, day, hour, minute, second


def day_of_week(jd: float) -> int:
    """Finds the day of the week for a given Julian date.

    References:
        Vallado: 2022, p. 183, Eq. 3-42

    Args:
        jd (float): Julian date (days from 4713 BC)

    Returns:
        int: Day of the week (1 for Sunday, 2 for Monday, etc.)
    """
    # Ensure the Julian date corresponds to 0.0 hours of the day
    jd = int(jd + 0.5)

    # Calculate the day of the week (1 = Sunday, ..., 7 = Saturday)
    return ((jd + 1) % 7) + 1


def find_dst_date(year: int, month: int, target_week: int, weekday: int) -> int | None:
    """Find the nth occurrence or the last occurrence of a specific weekday in a given
    month.

    Args:
        year (int): Year of interest
        month (int): Month of interest
        target_week (int): Week index (1-based for nth week, -1 for last week)
        weekday (int): Day of the week (0=Monday, 6=Sunday)

    Returns:
        int: Day of the month for the specified weekday and week (None if invalid)
    """
    weeks = calendar.monthcalendar(year, month)
    if target_week == -1:  # last occurrence
        for week in reversed(weeks):
            if week[weekday] != 0:  # non-zero means the day exists
                return week[weekday]
    else:  # nth occurrence
        count = 0
        for week in weeks:
            if week[weekday] != 0:  # non-zero means the day exists
                count += 1
                if count == target_week:
                    return week[weekday]
    return None


def daylight_savings(year: int, lon: float) -> Tuple[int, int, float, float]:
    """Find the start and stop dates for daylight savings time (DST) in a given year.

    References:
        Vallado: 2022, p. 183

    This function uses U.S.-specific DST rules:
        - Before 2007: DST starts on the first Sunday of April and ends on the last
          Sunday of October.
        - From 2007 onward: DST starts on the second Sunday of March and ends on the
          first Sunday of November.

    Args:
        year (int): The year (1900 .. 2100)
        lon (float): Longitude of the site (WEST is negative) in radians

    Returns:
        tuple: (startday, stopday, jdstartdst, jdstopdst)
            startday (int): Day in March when DST starts
            stopday (int): Day in November when DST ends
            jdstartdst (float): Julian date of DST start
            jdstopdst (float): Julian date of DST end
    """
    # Check if the longitude corresponds to a U.S. time zone
    lon_deg = np.degrees(lon)
    if lon_deg < -125 or lon_deg > -66:
        logger.warning(
            "Longitude does not correspond to a U.S. time zone. DST dates may not align"
            "with local transitions."
        )

    # Determine the site zone (0.0 gives Greenwich time)
    zone = int(lon_deg / const.DEG2HR)
    if zone > 0:
        zone -= const.DAY2HR

    # Find the start and stop days for DST
    if year < DST_RULE_CHANGE_YEAR:
        # Pre-2007: DST starts 1st Sunday in April, ends last Sunday in October
        startday = find_dst_date(year, 4, 1, calendar.SUNDAY)
        stopday = find_dst_date(year, 10, -1, calendar.SUNDAY)
    else:
        # 2007 and later: DST starts 2nd Sunday in March, ends 1st Sunday in November
        startday = find_dst_date(year, 3, 2, calendar.SUNDAY)
        stopday = find_dst_date(year, 11, 1, calendar.SUNDAY)

    # Check for invalid dates
    if startday is None or stopday is None:
        logger.warning("Invalid DST dates. Using default values.")
        startday = 1 if year < DST_RULE_CHANGE_YEAR else 8
        stopday = 31

    # Julian date for DST start
    jdstartdst, jdstartdst_frac = jday(
        year, 3 if year >= DST_RULE_CHANGE_YEAR else 4, startday, 12, 0, 0
    )
    jdstartdst += jdstartdst_frac
    jdstartdst += (DST_CHANGE_UTC_HOUR - zone) / const.DAY2HR

    # Julian date for DST end
    jdstopdst, jdstopdst_frac = jday(
        year, 11 if year >= DST_RULE_CHANGE_YEAR else 10, stopday, 12, 0, 0
    )
    jdstopdst += jdstopdst_frac
    jdstopdst += (DST_CHANGE_UTC_HOUR - zone) / const.DAY2HR

    return startday, stopday, jdstartdst, jdstopdst


def convtime(
    year: int,
    month: int,
    day: int,
    hour: int,
    minute: int,
    second: float,
    timezone: int,
    dut1: float,
    dat: float,
) -> Tuple[
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
    float,
]:
    """Convert UTC or UT1 to various time systems.

    This function finds the time parameters and Julian Century values for inputs of UTC
    or UT1. Since calucations are in UTC, you must include timezone if you enter a local
    time; otherwise it should be zero.

    References:
        Vallado: 2022, p. 196, Algorithm 16

    Args:
        year (int): Year (1900 .. 2100)
        month (int): Month (1 .. 12)
        day (int): Day (1 .. 31)
        hour (int): Universal Time hour (0 .. 23)
        minute (int): Universal Time minute (0 .. 59)
        second (float): Universal Time second (0.0 .. 59.999)
        timezone (int): Offset to UTC from local time (0 .. 23 hr)
        dut1 (float): Delta of UT1 - UTC in seconds
        dat (float): Delta of TAI - UTC in seconds

    Returns:
        tuple: (ut1, tut1, jdut1, jdut1frac, utc, tai, tt, ttt, jdtt, jdttfrac, tdb,
                ttdb, jdtdb, jdtdbfrac)
            ut1 (float): UT1 in seconds (from start of day)
            tut1 (float): Julian centuries since J2000 for UT1
            jdut1 (float): Julian Date for UT1 (days only, days from 4713 BC)
            jdut1frac (float): Julian Date for UT1 (fractional part)
            utc (float): Coordinated Universal Time in seconds
            tai (float): International Atomic Time in seconds
            tt (float): Terrestrial Time in seconds (from start of day)
            ttt (float): Julian centuries since J2000 for TT
            jdtt (float): Julian Date for TT (days only, days from 4713 BC)
            jdttfrac (float): Julian Date for TT (fractional part)
            tdb (float): Terrestrial Barycentric Time in seconds (from start of day)
            ttdb (float): Julian centuries since J2000 for TDB
            jdtdb (float): Julian Date for TDB (days only, days from 4713 BC)
            jdtdbfrac (float): Julian Date for TDB (fractional part)
    """
    # UTC in seconds from the start of the day
    local_hour = timezone + hour
    utc = hms2sec(local_hour, minute, second)

    # UT1
    ut1 = utc + dut1
    hrtemp, mintemp, sectemp = sec2hms(ut1)
    jdut1, jdut1frac = jday(year, month, day, hrtemp, mintemp, sectemp)
    tut1 = (jdut1 + jdut1frac - const.J2000) / const.CENT2DAY

    # TAI
    tai = utc + dat

    # TT
    tt = tai + 32.184  # seconds
    hrtemp, mintemp, sectemp = sec2hms(tt)
    jdtt, jdttfrac = jday(year, month, day, hrtemp, mintemp, sectemp)
    ttt = (jdtt + jdttfrac - const.J2000) / const.CENT2DAY

    # TDB
    tdb = (
        tt
        + 0.001657 * np.sin(628.3076 * ttt + 6.2401)
        + 0.000022 * np.sin(575.3385 * ttt + 4.297)
        + 0.000014 * np.sin(1256.6152 * ttt + 6.1969)
        + 0.000005 * np.sin(606.9777 * ttt + 4.0212)
        + 0.000005 * np.sin(52.9691 * ttt + 0.4444)
        + 0.000002 * np.sin(21.3299 * ttt + 5.5431)
        + 0.00001 * ttt * np.sin(628.3076 * ttt + 4.249)
    )
    hrtemp, mintemp, sectemp = sec2hms(tdb)
    jdtdb, jdtdbfrac = jday(year, month, day, hrtemp, mintemp, sectemp)
    ttdb = (jdtdb + jdtdbfrac - const.J2000) / const.CENT2DAY

    return (
        ut1,
        tut1,
        jdut1,
        jdut1frac,
        utc,
        tai,
        tt,
        ttt,
        jdtt,
        jdttfrac,
        tdb,
        ttdb,
        jdtdb,
        jdtdbfrac,
    )
